#include "checkfeas.h"

extern int *roomSize, *eventSize, *totalNumConflict, **before, *originalTotalNumPlaces;
extern bool **attends, **roomFeatures, **eventRoom, **conflict, **eventFeatures, **originalEventPlace, **eventAvail, **roomAvail, *hasPrecConstr;
extern int numEvents, numRooms, numFeatures, numStudents, NUMBEROFPLACES, numDummyRooms;
extern bool outputOn;
extern TwoDIntVector precidence, beforePrec, feasRooms, eventStus;


//------------------------------------------------------------------------------
inline
int doIndependentSoftCostCalculation(TTAndUnplaced &theSolution)
{
	//Does an independent calculation of the soft cost. First generate the relevant matrix	
	bool** stuavail = new bool* [NUMBEROFSLOTS];
	int i,r,c,s,day,totalCost=0;;
	for (i=0;i<NUMBEROFSLOTS;i++) stuavail[i] = new bool[numStudents];
	for (r=0;r<NUMBEROFSLOTS;r++) for (c=0;c<numStudents;c++) stuavail[r][c] = true;
	for (r=0;r<numRooms;r++){
		for (c=0; c<NUMBEROFSLOTS; c++){
			if (theSolution.TT[r][c] != -1){
				//Mark a false every student that attends this event in the relevant slot
				for (i=0; i<numStudents; i++) if (attends[i][theSolution.TT[r][c]]) stuavail[c][i] = false;
			}
		}
	}
	//Now generate the cost
	for(day = 0; day<NUMDAYS; day++){
		for(s = 0; s<numStudents; s++){
			int longintensive = 0, single = 0, lastslot = 0, licount = 0, scount = 0;
			for (r=day*SLOTSPERDAY; r<(day*SLOTSPERDAY)+SLOTSPERDAY; r++){
				if (!stuavail[r][s]) licount++;
				else licount = 0;
				if (licount>=3) longintensive++;
				if (!stuavail[r][s]) scount++;
			}
			if (scount == 1) single++;
			if (!stuavail[(day*SLOTSPERDAY)+LASTSLOTINDEX][s]) lastslot++;
			totalCost += longintensive+lastslot+single;
		}
	}
	//Delete the arrays
	for (i=0;i<NUMBEROFSLOTS;i++) delete [] stuavail[i];
	delete [] stuavail;
	return totalCost;
}
	

//----------------------------------------------------------------------------------------
void checkFeasibility(TTAndUnplaced &theSolution)
{
	if(outputOn)cout<<"\n\n";
	//This simply checks for feasibility
	bool valid = true;
	bool feasible = true;
	int r,c,i,row;
		
	//1) Check that there are no unplaced events in the unplaced list
	if(theSolution.Unplaced.size()>0){
		feasible = false;
		if(outputOn){
			cout<<"There are "<<theSolution.Unplaced.size()<<" unplaced events in the unplaced list. These are.. {";
			for(i=0;i<theSolution.Unplaced.size()-1;i++)
				cout<<theSolution.Unplaced[i]<<", ";
			cout<<theSolution.Unplaced[theSolution.Unplaced.size()-1]<<"}\n";
		}
	}

	
	//Check that all events are present in the solution somewhere.
	int* evs = new int [numEvents];
	for(i = 0; i<numEvents; i++)evs[i]=0;
	for(r=0;r<numRooms;r++){
		for(c=0;c<NUMBEROFSLOTS;c++){
			if(theSolution.TT[r][c] != -1) 
				evs[theSolution.TT[r][c]]++;
		}
	}
	for(i=0;i<theSolution.Unplaced.size();i++){
		evs[theSolution.Unplaced[i]]++;
	}
	for(i = 0; i<numEvents; i++){
		if(evs[i]!=1){
			valid = false;
			feasible = false;
			cout << "Error: Event "<<i<<" occurs "<<evs[i]<<" times in the timetable/unplaced list"<<endl;
		}
	}
	delete[]evs;
		
	//2 Check all events are in feasible rooms
	for(c=0;c<NUMBEROFSLOTS;c++){
		for(r=0;r<numRooms;r++){
			if(theSolution.TT[r][c] != -1){
				if( ! eventRoom[theSolution.TT[r][c]][r]){
					valid = false;
					feasible = false;
					cout << "Error. Cannot put event "<<theSolution.TT[r][c]<<" in room "<<r<<endl;
				}
			}
		}
	}
	//3) Check all columns are feasible
	for(c=0;c<NUMBEROFSLOTS;c++){
		for(row = 0; row<numRooms-1; row++){
			for(r=row+1; r<numRooms; r++){
				if(theSolution.TT[r][c] != -1 && theSolution.TT[row][c] != -1){
					if(conflict[theSolution.TT[r][c]][theSolution.TT[row][c]]){
						valid = false;
						feasible = false;
						cout << "Error. Events "<<theSolution.TT[r][c]<<" and "<<theSolution.TT[row][c]<<" conflict in column "<<c<<endl;
					}
				}
			}
		}
	}
	//4) Check all events are in feasible timeslots
	for(c=0;c<NUMBEROFSLOTS;c++){
		for(r=0;r<numRooms;r++){
			if(theSolution.TT[r][c] != -1){
				if(!eventAvail[theSolution.TT[r][c]][c]){
					valid = false;
					feasible = false;
					cout<< "Error. Event "<<theSolution.TT[r][c]<<" is not supposed to go into slot "<<c<<endl;
				}
			}
		}
	}
	//5) Check all events are in timeslots where the rooms are available
	for(c=0;c<NUMBEROFSLOTS;c++){
		for(r=0;r<numRooms;r++){
			if(theSolution.TT[r][c] != -1){
				if(!roomAvail[r][c]){
					valid = false;
					feasible = false;
					cout<< "Error. Event "<<theSolution.TT[r][c]<<" is placed in slot "<<c<<" and room "<<r<<". This room is not available here"<<endl;
				}
			}
		}
	}
	
	int event1, c1, r1, event2;
	int precidenceViolations = 0;
	for(c=0; c<NUMBEROFSLOTS-1; c++){
		for(r=0; r<numRooms; r++){
			event1 = theSolution.TT[r][c];
			if(event1 != -1){
				//Now go through every timeslot after c and check whether there is an event that should have gone after event_1
				for(c1 = c; c1<NUMBEROFSLOTS; c1++){
					for(r1=0; r1<numRooms; r1++){
						event2 = theSolution.TT[r1][c1];
						if(event1 != event2 && event2 != -1){
							if(before[event2][event1] == 1){
								valid = false;
								feasible = false;
								cout<<"Error ("<<precidenceViolations<<"): Event "<<event2<<" has been scheuled after event "<<event1<<". It is in timeslot "<<c1<<", which is after timeslot "<<c<<" where "<<event1<<" is scheduled\n";
								precidenceViolations++;
							}
						}
					}
				}
			}
		}
	}
	
	//Now do an independent calculation of the soft cost. 	
	if(outputOn) cout<<"Total Cost Check = "<<doIndependentSoftCostCalculation(theSolution)<<"\n";
	
	if(!valid){
		cout<<"Error timetable is not valid as it breaks some hard constraints."<<endl;
		exit(1);
	}
	if(!feasible && valid){
		if(outputOn) cout<<"Timetable is valid but not feasible (|U| = "<<theSolution.Unplaced.size()<<")"<<endl;
	}
	else //(feasible && valid) 
		if(outputOn) cout<<"The timetable is valid and fully feasible. All events are placed"<<endl;
}


void eraseDummyRooms(TTAndUnplaced &theSolution, int &softCost)
{
	//Reset the number of rooms to the correct amount
	numRooms = numRooms - numDummyRooms;
	int i, j;

	//Make the empty timetable and copy in the information
	TTAndUnplaced newSol;
	vector<vector<int> > temp(numRooms, vector<int>(NUMBEROFSLOTS));
	newSol.TT.swap(temp);
	for(i=0;i<numRooms;i++){
		for(j=0;j<NUMBEROFSLOTS;j++){
			newSol.TT[i][j] = theSolution.TT[i][j];
		}
	}
	for(i=numRooms;i<numRooms+numDummyRooms;i++){
		for(j=0;j<NUMBEROFSLOTS;j++){
			if(theSolution.TT[i][j] != -1) newSol.Unplaced.push_back(theSolution.TT[i][j]);
		}
	}
	if(outputOn) cout<<"There were "<<newSol.Unplaced.size()<<" events in the dummy rooms that have now been removed\n";
	softCost = doIndependentSoftCostCalculation(theSolution);
	theSolution.TT.swap(newSol.TT);
	theSolution.Unplaced.swap(newSol.Unplaced);
}
