#include "initsolgenerator.h"

extern int *roomSize, *eventSize, *totalNumConflict, **before, *originalTotalNumPlaces;
extern bool **attends, **roomFeatures, **eventRoom, **conflict, **eventFeatures, **originalEventPlace, **eventAvail, **roomAvail, *hasPrecConstr;
extern int numEvents, numRooms, numFeatures, numStudents, NUMBEROFPLACES;
extern bool outputOn;
extern TwoDIntVector precidence, beforePrec, feasRooms, eventStus;

//-----------------------------------------------------------------------------------
TTAndUnplaced generateNewSolution()
{
	if(outputOn)cout<<"\nMaking Initial Solution... "<<endl;
	int r, i, c, x, y, slot, room;
	
	//MAKE-A-COPY-OF-EVENT-PLACE-MATRIX-and-the total num places. These are used locally within the construction process.
	bool** eventPlace = new bool* [numEvents];
	for (r=0; r<numEvents; r++) eventPlace[r] = new bool [NUMBEROFPLACES];
	for (r=0; r< numEvents; r++) for (c=0; c<NUMBEROFPLACES; c++) eventPlace[r][c] = originalEventPlace[r][c];
	int* numPossPlace = new int [numEvents];
	for(r=0;r<numEvents;r++) numPossPlace[r] = originalTotalNumPlaces[r];

	//Also make an array signifying how many events can go in each place
	int* numEvsThatCanGoInPlace = new int [NUMBEROFPLACES];
	for(c=0;c<NUMBEROFPLACES;c++){
		int pCount = 0;
		for(r=0;r<numEvents;r++)if(eventPlace[r][c])pCount++;
		numEvsThatCanGoInPlace[c]=pCount;
	}

	//Make the empty the timetable
	TTAndUnplaced theSolution;
	vector<vector<int> > temp(numRooms, vector<int>(NUMBEROFSLOTS,-1));
	theSolution.TT.swap(temp);
	
	//we now wish to select an event and a place for it
	int evsPlaced = 0;
	int numEventsWithPlacesLeft = numEvents;
	while(numEventsWithPlacesLeft > 0){
		x = select_event(numPossPlace);
		y = select_place(x, eventPlace, theSolution, numEvsThatCanGoInPlace);
		slot = y/numRooms;
		room = y%numRooms;
		theSolution.TT[room][slot]=x;
		evsPlaced++;
		//We now update the data structure to reflect the changes
		updateMatricies(eventPlace, numPossPlace, x, y, numEventsWithPlacesLeft, numEvsThatCanGoInPlace, beforePrec);
	}//end of while
	
	if(outputOn)cout<<"...Done. ("<<evsPlaced<<"/"<<numEvents<<" events placed)\n";

	//When we get here, we have a feasible timetable and possibly some unplaced events
	//We now need to identify which events still need to be placed (these will be the ones in marked
	//by a zero in the numPossPlace array (placed evs will be marked by a -1), put these into an array
	//and pass this array to the procedure which opens up extra slots
	for(i=0;i<numEvents;i++){
		if(numPossPlace[i]==0){ //i.e we haven't placed event i
			theSolution.Unplaced.push_back(i);
		}
	}
		
	//Delete the arrays we've constructed in this function
	for (i=0; i<numEvents; i++) delete [] eventPlace[i];
	delete [] eventPlace;
	delete [] numPossPlace;
	delete [] numEvsThatCanGoInPlace;
	
	return(theSolution);
}

//-----------------------------------------------------------------
inline
int select_event(int* numPossPlace)
{
	//This is the select event process. The function ignores
	//any event that has -1 numPossPlaces (i.e. it has been placed) and 0 numPossPlaces (i.e
	//it hasn't been placed and theres no place for it). It also returns the POSITION
	//of the event in the numPossPlace Array (not the event identifyer itself)
	int minimum = INT_MAX;
	int i;
	IntVector min_vals;
	for (i=0; i<numEvents; i++){
		if (numPossPlace[i] != -1 && numPossPlace[i] != 0){	
			if (numPossPlace[i] < minimum){
				min_vals.clear();
				min_vals.push_back(i);
				minimum = numPossPlace[i];
			}
			else if (numPossPlace[i] == minimum){
				min_vals.push_back(i);
			}
			else; //do nothing
			
		}
	}
	//We now will have a list indicating all the events which have
	//the least possible places so we select one and return.
	return(chooseEventWithMostConflicts(min_vals, totalNumConflict));	
}

//-----------------------------------------------------------------
inline
int chooseEventWithMostConflicts(IntVector the_list, int *totalNumConflicts)
{
	//This heuristic chooses the most "conflicting" event out of the elemnts in the list. 
	//If there is more than one with maximal size, choose between these at random
	IntVector maxList;
	int theElement, the_iter;
	int maxVal = -1;
	for(the_iter = 0; the_iter < the_list.size(); the_iter++){
		theElement = the_list[the_iter];
		if(totalNumConflicts[theElement] > maxVal){
			maxList.clear();
			maxList.push_back(theElement);
			maxVal = totalNumConflicts[theElement];
		}
		else if (totalNumConflicts[theElement] == maxVal){
			maxList.push_back(theElement);
		}
		else;//do nothing
	}
	//We now have a list of at least one element that have the largest degree
	//we select one of these and end
	int r = rand() % maxList.size();
	return(maxList[r]);
}

//-----------------------------------------------------------------
inline
int select_place(int event, bool** event_place, TTAndUnplaced &theSolution, int *numEvsThatCanGoInPlace)
{
	int minimumq1 = INT_MAX;
	int c, q1;
	IntVector min_vals;
	for (c=0; c<NUMBEROFPLACES; c++){
		if (event_place[event][c]){
			//work out how many other events could go in this place (minus 1 because we are considering other events
			q1 = numEvsThatCanGoInPlace[c]-1;
			if(q1<0){
				cout<<"Error in select_place "<<endl;
				exit(1);
			}

			//IF THIS VALUE IS THE LOWEST SO FAR, OR EQUAL TO THE LOWEST, WE STICK IT IN A LIST.
			if(q1 < minimumq1){
				min_vals.clear();
				min_vals.push_back(c);
				minimumq1 = q1;
			}
			else if (q1 == minimumq1){
				min_vals.push_back(c);
			}
			else; //do nothing
	
		}//end of if
	}//end of for loop
	//We now have at least one event in a list called min_vals. If there are more than
	//one in the list then we need to choose one. We do this using the second crieria
	//i.e. we are using SEQUENTIAL EVALUATION
	return(chooseElWithMostFullSlot(min_vals, theSolution));
}

//--------------------------------------------------------------------------
inline
int chooseElWithMostFullSlot(IntVector the_list, TTAndUnplaced &theSolution)
{
	IntVector maxList;
	int q2, theElement, the_iter;
	int maximumQ2 = -1;
	for(the_iter = 0; the_iter < the_list.size(); the_iter++){
		
		theElement = the_list[the_iter];
		int slot = theElement / numRooms;
		int	count = 0;
		for (int i=0; i<numRooms; i++)			
			if (theSolution.TT[i][slot] != -1) count++;
		q2 = count;
		
		if(q2>maximumQ2){
			maxList.clear();
			maxList.push_back(theElement);
			maximumQ2 = q2;
		}
		
		else if(q2==maximumQ2){
			maxList.push_back(theElement);
		}
		else; //do nothing
	}
	//We now have a list of at least one element we select one of these and end
	int r = rand() % maxList.size();
	return maxList[r];
}

//--------------------------------------------------------------------------
inline
void updateMatricies(bool **eventPlace, int *numPossPlace, int x, int y, int &eventsWithPlacesLeft, int *numEvsThatCanGoInPlace,
					TwoDIntVector &beforePrec)
{
	// x = the event we just inserted
	// y = the place where we just inserted it
	
	//Any event that could go into place y now cannot go there. Thus they lose a place option
	int i, r, c;
	for (i=0;i<x;i++){
		if (eventPlace[i][y]){
			eventPlace[i][y] = false;
			numEvsThatCanGoInPlace[y]--;
			numPossPlace[i]--;
			if(numPossPlace[i]==0)eventsWithPlacesLeft--;
		}
	}
	for (i=x+1;i<numEvents;i++){
		if (eventPlace[i][y]){
			eventPlace[i][y] = false;
			numEvsThatCanGoInPlace[y]--;
			numPossPlace[i]--;
			if(numPossPlace[i]==0)eventsWithPlacesLeft--;
		}
	}

	//change all true entries of event x's row to false and update the other arrays too
	for (i=0;i<NUMBEROFPLACES;i++){
		if(eventPlace[x][i]){
			eventPlace[x][i] = false;
			numEvsThatCanGoInPlace[i]--;
		}
	}
	//change x's entry to -1 to signify that it has been placed 
	numPossPlace[x] = -1;
	eventsWithPlacesLeft--;

	//For each event r, if r and x conflict, then all the all of r's places in timeslot y/numRooms
	//(i.e where x has just been inserted) need to be turned to false too
	int slot = y/numRooms, precEvent;
	for(c=slot*numRooms; c<(slot+1)*numRooms; c++){
		for(r=0;r<numEvents;r++){
			if(numPossPlace[r] <= 0){
				;//do nothing - ignore the row r
			} else if (conflict[x][r] && eventPlace[r][c]){
				eventPlace[r][c] = false;
				numEvsThatCanGoInPlace[c]--;
				numPossPlace[r]--;
				if(numPossPlace[r]==0)eventsWithPlacesLeft--;
			}
		}
	}

	//If x has any events that have to come after it, then mark all timeslots to the left as unavailable for these events.
	for(r=0;r<precidence[x].size();r++){
		precEvent = precidence[x][r];
		if(numPossPlace[precEvent] <= 0){
			;//do nothing - precEvent has already been placed or has no available places
		} else {
			//precEvent has not yet been placed. We now know it cannot be placed into timeslots 0 to "slot"
			for(c=0; c<(slot+1)*numRooms; c++){
				if(eventPlace[precEvent][c]){
					eventPlace[precEvent][c] = false;
					numEvsThatCanGoInPlace[c]--;
					numPossPlace[precEvent]--;
					if(numPossPlace[precEvent]==0)eventsWithPlacesLeft--;
				}
			}
		}
	}

	//Similarly x has any events that have to before, then mark all relevant timeslots to the right as unavailable for these events.
	for(r=0;r<beforePrec[x].size();r++){
		precEvent = beforePrec[x][r];
		if(numPossPlace[precEvent] <= 0){
			;//do nothing - precEvent has already been placed or has no available places
		} else {
			//precEvent has not yet been placed. We now know it cannot be placed into timeslots 0 to "slot"
			for(c=slot*numRooms; c<NUMBEROFPLACES; c++){
				if(eventPlace[precEvent][c]){
					eventPlace[precEvent][c] = false;
					numEvsThatCanGoInPlace[c]--;
					numPossPlace[precEvent]--;
					if(numPossPlace[precEvent]==0)eventsWithPlacesLeft--;
				}
			}
		}
	}
}


