#include "readinputandmakematricies.h"

extern int *roomSize, *eventSize, *totalNumConflict, **before, *originalTotalNumPlaces;
extern bool **attends, **roomFeatures, **eventRoom, **conflict, **eventFeatures, **originalEventPlace, **eventAvail, **roomAvail, *hasPrecConstr;
extern int numEvents, numRooms, numFeatures, numStudents, NUMBEROFPLACES, numDummyRooms;
extern TwoDIntVector precidence, beforePrec, feasRooms, eventStus;
extern bool outputOn;

//-------------------------------------------------------------------------------------
void readTimFileAndPreProcess(ifstream& inStream, bool legacyTim)
{
	//Read in the first line of the tim file
	inStream >> numEvents >> numRooms >> numFeatures >> numStudents;

	roomSize = makeRoomSizeMatrix(inStream);
	attends  = makeAttendsMatrix(inStream);
	roomFeatures = makeRoomFeaturesMatrix(inStream);
	eventFeatures = makeEventFeaturesMatrix(inStream);
	eventAvail = makeEventAvail(inStream, legacyTim);

	//Having read in all the stuff from the files to do with rooms, we now add the dummy rooms (if required)
	numRooms = numRooms + numDummyRooms;
	NUMBEROFPLACES = numRooms*NUMBEROFSLOTS;
	
	//This does nothing except make everything true (becasue room availability is not considered in the tt competiton specifications
	roomAvail = makeRoomAvail(inStream, legacyTim);
	
	//Finally, read in the precidence constraints
	before = makeBeforeMatrix(inStream, legacyTim);
	
	//Now we have finished reading in the file, and go on to do some pre-processing.
	makeEventSizeMatrix();
	makeEventRoomMatrix();
	makeEventConflictMatrix();
	makeTotalNumConflict();
	originalEventPlace = makeEventPlaceMatrix();
	originalTotalNumPlaces = makeTotalNumPlaces();
	makePrecidence();
	modifyConflictsForPrecidence();
	makeEventStus();
}


//-------------------------------------------------------------------------------------
int* makeRoomSizeMatrix(ifstream& i_stream){
	int i;
	int* r_size;
	r_size = new int[numRooms+numDummyRooms];
	for (i = 0; i < numRooms; i++) i_stream >> r_size[i];
	//If we are using any dummy rooms, make their sizes infinite
	if(numDummyRooms > 0){
		for (i=numRooms; i<numRooms+numDummyRooms; i++) r_size[i] = INT_MAX;
	}
	return (r_size);
}

//-------------------------------------------------------------------------------------
bool** makeAttendsMatrix(ifstream& i_stream){  
	int i, r, c;
	bool** att_matrix;
	att_matrix = new bool* [numStudents];
	for (i = 0; i<numStudents; i++)	att_matrix[i] = new bool [numEvents];
	for (r = 0; r < numStudents; r++){
		for (c = 0; c < numEvents; c++){
			i_stream >> i;
			if(i==1)att_matrix[r][c] = true;
			else att_matrix[r][c] = false;
		}
	}
	return (att_matrix);
}

//-------------------------------------------------------------------------------------
bool** makeRoomFeaturesMatrix(ifstream& i_stream){
	int r, c, i;
	bool** r_features;
	r_features = new bool* [numRooms+numDummyRooms];
	for (i = 0; i<numRooms+numDummyRooms; i++)r_features[i] = new bool [numFeatures];
	for (r = 0; r < numRooms; r++){
		for (c = 0; c < numFeatures; c++){
			i_stream >> i;
			if(i==1)r_features[r][c]=true;
			else r_features[r][c]=false;
		}
	}
	//If we are using any dummy rooms, allocate them all of the features
	if(numDummyRooms > 0){
		for(r=numRooms; r<numRooms+numDummyRooms; r++)
			for(c=0;c<numFeatures;c++)
				r_features[r][c]=true;
	}
	return (r_features);
}

//-------------------------------------------------------------------------------------
bool** makeEventFeaturesMatrix(ifstream& i_stream){
	int r, c, i;
	bool** e_features;
	e_features = new bool* [numEvents];
	for (i=0; i <numEvents; i++) e_features[i] = new bool [numFeatures];
	for (r = 0; r < numEvents; r++){
		for (c=0; c < numFeatures; c++){
			i_stream >> i;
			if(i==1)e_features[r][c]=true;
			else e_features[r][c]=false;
		}
	}
	return (e_features);
}


//-------------------------------------------------------------------------------------
bool** makeEventAvail(ifstream& i_stream, bool legacyTim)
{
	int r, c, i;
	bool** evAvail;
	evAvail = new bool* [numEvents];
	for (i=0; i <numEvents; i++) evAvail[i] = new bool [NUMBEROFSLOTS];
	if(legacyTim){
		//We are reading an old tim file so there are no eventAvail constraints
		for(r=0;r<numEvents;r++)for(c=0;c<NUMBEROFSLOTS;c++)evAvail[r][c]=true;
	}
	else {
		for (r = 0; r < numEvents; r++){
			for (c=0; c < NUMBEROFSLOTS; c++){
				i_stream >> i;
				if(i==1) evAvail[r][c]=true;
				else evAvail[r][c]=false;
			}
		}
	}
	return (evAvail);
}

//-------------------------------------------------------------------------------------
bool** makeRoomAvail(ifstream& i_stream, bool legacyTim)
{
	int r, c, i;
	bool** rmAvail;
	rmAvail = new bool* [numRooms];
	for(i=0; i<numRooms; i++)rmAvail[i] = new bool [NUMBEROFSLOTS];
	for(r=0; r<numRooms; r++)for(c=0; c<NUMBEROFSLOTS; c++)rmAvail[r][c]=true;
	return (rmAvail);
}

//-------------------------------------------------------------------------------------
int** makeBeforeMatrix(ifstream &i_stream, bool legacyTim)
{
	int r, c, i;
	int** before;
	before = new int* [numEvents];
	for (i=0; i <numEvents; i++)before[i] = new int [numEvents];
	if(legacyTim){
		//We are reading an old tim file so there are no precidence constraints
		for(r=0;r<numEvents;r++)for(c=0;c<numEvents;c++)before[r][c]=0;
	}
	else {
		for (r = 0; r < numEvents; r++){
			for (c=0; c < numEvents; c++){
				i_stream >> before[r][c];
			}
		}
	}
	return (before);
}

//-------------------------------------------------------------------------------------
void makeEventSizeMatrix(){
	int r, c, total;
	eventSize = new int[numEvents];
	for (c = 0; c < numEvents; c++){
		total = 0;
		for (r = 0; r < numStudents; r++) if (attends[r][c] == 1) total++;
		eventSize[c] = total;
	}
}

//-------------------------------------------------------------------------------------
void makeEventRoomMatrix()
{
	int i, r, c;
    eventRoom = new bool* [numEvents];
	for (i=0; i < numEvents; i++) eventRoom[i] = new bool [numRooms];
	//We step through each cell in turn in the matrix and we check to see
	//that if event r demands it, room c supplies it.
	for (r = 0; r < numEvents; r++){
		for (c = 0; c < numRooms; c++){
			int count = 0;
			int suitable = 1; 
			while ((count < numFeatures) && (suitable == 1)){
				if (eventFeatures[r][count] == 0){
					//event r doesnt require the feature - we dont care
					count++;
				}
				else {	
					//The event	must have a requirement
					if (roomFeatures[c][count] == 1){
						//But its OK because room c supplies it
						count++;
					}
					else{
						//The room doesnt satisfy the requirement
						suitable = 0;
					}
				}
			}
			if (suitable == 1)	eventRoom[r][c] = true;
			else eventRoom[r][c] = false;
		}
	}
		
	//Now we modify the event room matrix using event_size array
	for (r = 0; r < numEvents; r++){
		for (c = 0; c < numRooms; c++){
			if (eventSize[r] > roomSize[c]) eventRoom[r][c] = false;
		}
	}

	//For each event, calculate a list of the rooms it can go in.
	for(r=0;r<numEvents;r++){
		IntVector newIV;
		feasRooms.push_back(newIV);
		for(c=0;c<numRooms;c++){
			if(eventRoom[r][c]) feasRooms[r].push_back(c);
		}
	}
}

//-------------------------------------------------------------------------------------
void makeEventConflictMatrix()
{	
	int i, r, c;
	// PREPROCESSING STUFF 2 - the event_conflict matrix
	// This describes which events can be placed in which time slots together. A
	// zero says there is no conflict, a 1 says there is a conflict (and the 
	// corresponding events may not be placed in the same time slot)
	conflict = new bool* [numEvents];
	for (i=0; i < numEvents; i++)conflict[i] = new bool [numEvents];
	//We fill the matrix
	for (r = 0; r < numEvents; r++)	for (c = 0; c < numEvents; c++)	conflict[r][c] = false;
	//Now, for each cell in the matrix, check to see if there is a student who wishes to take the 2 corresponding events. if so we set the cell to 1
	for (r = 0; r < numEvents; r++){
		for (c = 0; c < numEvents; c++){
			bool clash = false;
			i = 0;
			while ((i < numStudents) && (!clash)){
				if ((attends[i][r] == true) && (attends[i][c] == true)){
					conflict[r][c] = true;
					clash = true;
				}
				else i++;
			}
		}
	}

	modifyForTightRooms();
	modiyForDifferentPreAssignments();
}


//-------------------------------------------------------------------------------------
void modifyForTightRooms()
{
	//This modifies the event_conflict matrix. We group together all rooms that are
	//only allowed to be scheduled in one particular room (we refer to them as 
	//'tight'. Obviosly 2+ events which are tight for the same room cannot be 
	//scheduled together
	int r,r1;
	for (r = 0; r < numEvents-1; r++){
		if (tight(r)){
			for (r1 = r+1; r1 < numEvents; r1++){
				if (tight(r1) && (tightRoom(r) == tightRoom(r1))){
					conflict[r][r1] = true;
					conflict[r1][r] = true;
				}
			}
		}
	}
}


//-------------------------------------------------------------------------------------
void modiyForDifferentPreAssignments()
{
	//This further modifies the conflicts matrix. If we have two events that
	//have been preassigned to mutually exclusive subsets of timeslots (e.g. event X can only go in
	//slots 3,4,and 5; and event Y can only go into 10,11,and 12, then obviously
	//no feasible timetable can have these in the same slots.
	int e1, e2;
	for (e1 = 0; e1 < numEvents-1; e1++){
		for (e2 = e1+1; e2 < numEvents; e2++){
			if(!conflict[e1][e2]){
				//There is no edge between e1 and e2 so we'll see if we can add one
				int c = 0;
				bool foundCommonSlot = false;
				while(c<NUMBEROFSLOTS && !foundCommonSlot){
					if(eventAvail[e1][c] && eventAvail[e2][c]){
						foundCommonSlot = true;
						break;
					}
					c++;
				}
				if(!foundCommonSlot){
					//if we are here, we know that there is no slot for which e1 and e2 are both available. 
					//We can therefore have an edge added to conflicts matrix
					conflict[e1][e2] = true;
					conflict[e2][e1] = true;
				}
			}
		}
	}
}

//-------------------------------------------------------------------------------------
void makeTotalNumConflict()
{
	//we construct a 1D array that tells us how many other events
	//each event conflicts with
	totalNumConflict = new int [numEvents];
	for(int r=0; r<numEvents; r++){
		int total = 0;
		for(int c=0; c<numEvents; c++) if(conflict[r][c])total++;
		totalNumConflict[r]=total-1;
	}
}

//-------------------------------------------------------------------------------------
bool** makeEventPlaceMatrix()
{
	int i, r, c;
    bool** eventPlace;
	eventPlace = new bool* [numEvents];
	for (i=0; i<numEvents; i++) eventPlace[i] = new bool [NUMBEROFPLACES];
	for (r=0; r<numEvents; r++){
		for (c=0; c<NUMBEROFPLACES; c++){
			int room = c%numRooms;
			int slot = c/numRooms;
			if (eventRoom[r][room] && eventAvail[r][slot] && roomAvail[room][slot]){
				//i.e. the "place" is ok for event r if and only if (1) the room is suitable
				//(2) theSlot is suitable (3) the room is available in the slot
				eventPlace[r][c] = true;
			}
			else eventPlace[r][c] = false;
		}
	}
	return (eventPlace);
}

//-------------------------------------------------------------------------------------
int* makeTotalNumPlaces()
{
	//we construct a 1D array that tells us how many places each event can go into
	//in the empty timetable
	int *totalNumPlaces = new int [numEvents];
	for(int r=0; r<numEvents; r++){
		int total = 0;
		for(int c=0; c<NUMBEROFPLACES; c++) if(originalEventPlace[r][c])total++;
		totalNumPlaces[r]=total;
	}
	return(totalNumPlaces);
}


//-------------------------------------------------------------------------------------
bool tight(int row){
	int count = 0;
	int col = 0;
	while ((col < numRooms) && (count < 2)){
		if (eventRoom[row][col])
			count = count + 1;
		col++;
	}
	if (count == 0){
		cout << "There appears to be an event with no suitable room in row "<< row<<endl;;
		exit(1);
	}
	else{
		if (count == 1) return true;
		else return false;
	}
}

//-------------------------------------------------------------------------------------
int tightRoom(int row)
{
	//We assume that rFoom row is tight
	int col = 0;
	bool found = false;
	while ((col < numRooms) && (found == false)){
		if (eventRoom[row][col] == true){
			found = true;
		}
		else col++;
	}
	if (found == false){
		cout << "There was an error in the tight room function";
		exit(1);
	}
	else return col;
}


//-------------------------------------------------------------------------------------
inline
bool addItemIfAppropriate(int theItem, int theRow)
{
	//Add theItem to theRow if its not there already
	int i;
	for(i=0; i<precidence[theRow].size(); i++){
		if(precidence[theRow][i] == theItem) return false;
	}
	//if we are here then we can add the item
	precidence[theRow].push_back(theItem);
	return true;
}
void makePrecidence()
{
	int r, i, c, j, k, x;
	//Make Precidence 2D Vector (i.e. a compressed version of the before matrix)
	for(i=0;i<numEvents;i++){
		IntVector newIV;
		precidence.push_back(newIV);
	}
	for(r=0;r<numEvents;r++){
		for(c=0;c<numEvents;c++){
			if(before[r][c]==1)	precidence[r].push_back(c);
		}
	}

	//We can now update the precidence constraints using the fact that the before replationship is transitive
	//i.e. IF (a -> b) and (b -> c) then (a -> c)   , where "->" means "should come before"
	bool addedAConstraint;
	for(i=0;i<numEvents;i++){
		for(j=0;j<precidence[i].size();j++){
			x = precidence[i][j];
			//we know that (i -> x). So now we identify what events y have (x -> y). These
			//y's are then added to i's row if needed
			for(k=0; k<precidence[x].size(); k++){
				addedAConstraint = addItemIfAppropriate(precidence[x][k],i);
				//if(addedAConstraint && outputOn) cout<<"Adding a precidence constraint due to the fact that ("<<i<<" -> "<<x<<") and ("<<x<<" -> "<<precidence[x][k]<<") and therefore ("<<i<<" -> "<<precidence[x][k]<<")"<<endl;
			}
		}
	}
	//Finally, also update the before matrix itself
	for(i=0;i<numEvents;i++){
		for(j=0;j<precidence[i].size();j++){
			x = precidence[i][j];
			before[i][x] = 1;
			before[x][i] = -1;
		}
	}

	//Also make a reverse precidence strucure: a list of events that must come before event i
	for(i=0;i<numEvents;i++){
		IntVector newIV;
		beforePrec.push_back(newIV);
	}
	for(r=0;r<numEvents;r++){
		for(c=0;c<numEvents;c++){
			if(before[r][c]==-1) beforePrec[r].push_back(c);
		}
	}

	//finally, make a 1d array that shows which events do and don't have associated precidence constraints
	hasPrecConstr = new bool[numEvents];
	for(i=0;i<numEvents;i++)hasPrecConstr[i]=false;
	for(i=0;i<numEvents;i++){
		if(!precidence[i].empty() || !beforePrec.empty()){
			for(j=0; j<precidence[i].size(); j++) 
				hasPrecConstr[precidence[i][j]]=true;
			hasPrecConstr[i]=true;
			for(j=0; j<beforePrec[i].size(); j++) 
				hasPrecConstr[beforePrec[i][j]]=true;
			hasPrecConstr[i]=true;
		}
	}
}

void modifyConflictsForPrecidence()
{
	//If we know that event a must be before event b, then a and b cannot go in the same slot and can therefore have an edge added between them.
	//Similarly, if we know that event a must be after event b, then a and b cannot go in the same slot and can therefore have an edge added between them.
	int i, j;
	for(i=0;i<numEvents;i++){
		for(j=0;j<precidence[i].size();j++){
			conflict[i][precidence[i][j]] = true;
			conflict[precidence[i][j]][i] = true;
		}
	}
	for(i=0;i<numEvents;i++){
		for(j=0;j<beforePrec[i].size();j++){
			conflict[i][beforePrec[i][j]] = true;
			conflict[beforePrec[i][j]][i] = true;
		}
	}
}

void makeEventStus(){
	//Make an adjacency list that lists the students taking each event (a compressed version of the attends matrix)
	int i,j;
	for(i=0;i<numEvents;i++){
		IntVector newIV;
		eventStus.push_back(newIV);
	}
	for(j=0;j<numEvents;j++){
		for(i=0;i<numStudents;i++){
			if(attends[i][j]) eventStus[j].push_back(i);
		}
	}
}

//-------------------------------------------------------------------------------------
void freeUpHeap()
{
	//We delete the arrays to free up the heap...
	int i;
	for (i=0; i<numStudents; i++)delete [] attends[i];
	delete [] attends;
	for (i=0; i<numRooms; i++)delete [] roomFeatures[i];
	delete [] roomFeatures;
	for (i=0; i<numEvents; i++)delete [] eventFeatures[i];
	delete [] eventFeatures;
	for (i=0; i<numEvents; i++)delete [] eventRoom[i];
	delete [] eventRoom;
	for (i=0; i<numEvents; i++)delete [] conflict[i];
	delete [] conflict;
	for (i=0; i<numEvents; i++) delete [] originalEventPlace[i];
	delete [] originalEventPlace;
	delete [] totalNumConflict;
	delete [] roomSize;
	delete [] eventSize;
	for (i=0; i<numEvents; i++)delete [] eventAvail[i];
	delete [] eventAvail;
	delete [] hasPrecConstr;
}

