#include "safunctions.h"

extern int *roomSize, *eventSize, *totalNumConflict, **before, *originalTotalNumPlaces;
extern bool **attends, **roomFeatures, **eventRoom, **conflict, **eventFeatures, **originalEventPlace, **eventAvail, **roomAvail, *hasPrecConstr;
extern int numEvents, numRooms, numFeatures, numStudents, NUMBEROFPLACES, numDummyRooms, weightCoefficient, maxNumKempeChains;
extern bool outputOn;
extern int flexibility;
extern IntVector NumKempeChainsRecord;
extern TwoDIntVector precidence, beforePrec, feasRooms, eventStus;
extern unsigned long long propMoves, feasMoves;

/****************************************************************************************/
/*******************        UNIVERSAL FUNCTIONS        **********************************/
/****************************************************************************************/
inline
void swapInts(int &a, int &b){
	int c = a; a = b; b = c;
}
inline 
double stDevP(int *a, int n)
{
	//First work out the average
	int total = 0, i;
	for(i=0; i<n; i++) total += a[i];
	double average = total/double(n);
	//Now calculate the variance (and then return the SD)
	double totalDiffsSquared=0;
	for(i=0; i<n; i++)
		totalDiffsSquared += pow((double(a[i])-average),2);
	return(sqrt(totalDiffsSquared/double(n-1)));
}
//------------------------------------------------------------------------------------
inline
int totalCost(int a[], int length)
{
	int total = 0, i;
	for (i=0; i<length; i++) total += a[i];
	return (total);
}
//------------------------------------------------------------------------------------
inline
int distToFeas(TTAndUnplaced &sol)
{
	int i, dist = 0;
	for(i=0; i<sol.Unplaced.size(); i++)dist += eventSize[sol.Unplaced[i]];
	return dist;
}
//------------------------------------------------------------------------------------
inline
void calcNewDummyCost(TTAndUnplaced &sol, int slot1, int slot2, IntVector &slot1Orig, IntVector &slot2Orig, int &newDummyCost, int &newEvsInDummy, int currentDummyCost, int currentEvsInDummy)
{
	if(numDummyRooms <= 0){
		newDummyCost = 0; newEvsInDummy = 0; return;
	}
	//Recalculates the dummy cost & num of events in dummy positions after a change has been made
	newDummyCost = currentDummyCost;
	newEvsInDummy = currentEvsInDummy;
	int r;
	for(r=numRooms-numDummyRooms;r<numRooms;r++){
		if(slot1Orig[r] != -1){
			newEvsInDummy--; 
			newDummyCost-=eventSize[slot1Orig[r]];
		}
		if(slot2Orig[r] != -1){
			newEvsInDummy--; 
			newDummyCost-=eventSize[slot2Orig[r]];
		}
		if(sol.TT[r][slot1] != -1){
			newEvsInDummy++; 
			newDummyCost+=eventSize[sol.TT[r][slot1]];
		}
		if(sol.TT[r][slot2] != -1){
			newEvsInDummy++; 
			newDummyCost+=eventSize[sol.TT[r][slot2]];
		}
	}
	return;
}
//------------------------------------------------------------------------------------
inline
void calcDummyCost(TTAndUnplaced &sol, int &dummyCost, int &evsInDummy)
{
	dummyCost=0;
	evsInDummy=0;
	if(numDummyRooms <= 0) return;
	//Calculates the dummy cost of the whole timetable by summing the sizes of the events in each dummy position
	int c, r;
	for(c=0;c<NUMBEROFSLOTS;c++){
		for(r=numRooms-numDummyRooms;r<numRooms;r++){
			if(sol.TT[r][c] != -1){
				dummyCost+=eventSize[sol.TT[r][c]];
				evsInDummy++;
			}
		}
	}
	return;
}
//------------------------------------------------------------------------------------
inline
int stuDayCost(int day, int theStudent, bool **studentavailability)
{
	//Calculate the score of the day for theStudent 
	int longintensive = 0, single = 0, lastslot = 0, licount = 0, scount = 0, r;
	for (r=day*SLOTSPERDAY; r<(day*SLOTSPERDAY)+SLOTSPERDAY; r++){
		if (!studentavailability[r][theStudent])
			licount++;
		else licount = 0;
		if (licount>=3)
			longintensive++;
		if (!studentavailability[r][theStudent])
			scount++;
	}
	if (scount == 1) 
		single++;
	if (!studentavailability[(day*SLOTSPERDAY)+LASTSLOTINDEX][theStudent]) 
		lastslot++;
	return(longintensive+lastslot+single);
}
//-----------------------------------------------------------------------------------
inline
void recalculateStudentTimetables(int ev1, int ev2, int slot1, int slot2, IntVector &d1Orig, IntVector &d2Orig, bool **studentavailability, int **stuDayCosts, int *dayCosts)
{
	//Calculates the cost change after a simple move (i.e. a move or a swap)
	int day1 = slot1/SLOTSPERDAY; int day2 = slot2/SLOTSPERDAY, c, i;
	if (day1 == day2){
		//We've only changed 1 day. We only need to recalculate the students involved in this
		if (ev1 != -1 && ev2 != -1){
			//Swapping two events
			for(c=0; c<numStudents; c++){
				if (attends[c][ev1] != attends[c][ev2]){
					//the student s is taking one of the events. We now alter the Savail and
					//sort the stuDayCosts out
					d1Orig[c] = stuDayCosts[day1][c]; 
					studentavailability[slot1][c] = !studentavailability[slot1][c];
					studentavailability[slot2][c] = !studentavailability[slot2][c];
					stuDayCosts[day1][c] = stuDayCost(day1,c,studentavailability);
					dayCosts[day1] = dayCosts[day1] + (stuDayCosts[day1][c] - d1Orig[c]);
				}
			}
		}
		else if (ev1 == -1 && ev2 != -1){ 
			//ev1 is a blank and ev2 is not (moving ev2)
			for(i=0; i<eventStus[ev2].size(); i++){
				c = eventStus[ev2][i];
				//student c is involved in the swap so we want to change his tt
				d1Orig[c] = stuDayCosts[day1][c];  
				studentavailability[slot1][c] = false; 
				studentavailability[slot2][c] = true;
				stuDayCosts[day1][c] = stuDayCost(day1,c,studentavailability);
				dayCosts[day1] = dayCosts[day1] + (stuDayCosts[day1][c] - d1Orig[c]);
			}
		}
		else if (ev1 != -1 && ev2 == -1){
			//ev1 is an event and ev2 is not (moving ev1)
			for(i=0; i<eventStus[ev1].size(); i++){
				c = eventStus[ev1][i];
				d1Orig[c] = stuDayCosts[day1][c]; 
				studentavailability[slot1][c] = true; 
				studentavailability[slot2][c] = false;
				stuDayCosts[day1][c] = stuDayCost(day1,c,studentavailability);
				dayCosts[day1] = dayCosts[day1] + (stuDayCosts[day1][c] - d1Orig[c]);
			}
		}
		else {
			cout << "Error in RecalculateStudentTimetables\n";
			exit(1);
		}
	}
	else {
		//Weve changed two days so need to recalculate them both
		if (ev1 != -1 && ev2 != -1){
			//Swapping two events
			for(c=0; c<numStudents; c++){
				if (attends[c][ev1] != attends[c][ev2]){
					//the student s is taking one of the events. We now alter the Savail and
					//sort the stuDayCosts out
					d1Orig[c] = stuDayCosts[day1][c]; 
					d2Orig[c] = stuDayCosts[day2][c]; 
					studentavailability[slot1][c] = !studentavailability[slot1][c];
					studentavailability[slot2][c] = !studentavailability[slot2][c];
					stuDayCosts[day1][c] = stuDayCost(day1,c,studentavailability);
					stuDayCosts[day2][c] = stuDayCost(day2,c,studentavailability);
					dayCosts[day1] = dayCosts[day1] + (stuDayCosts[day1][c] - d1Orig[c]);
					dayCosts[day2] = dayCosts[day2] + (stuDayCosts[day2][c] - d2Orig[c]);
				}
			}
		}
		else if (ev1 == -1 && ev2 != -1){
			//ev1 is a blank and ev2 is not
			for(i=0; i<eventStus[ev2].size(); i++){
				c = eventStus[ev2][i];
				//student c is involved in the swap so we want to change his tt
				d1Orig[c] = stuDayCosts[day1][c]; 
				d2Orig[c] = stuDayCosts[day2][c]; 
				studentavailability[slot1][c] = false; 
				studentavailability[slot2][c] = true;
				stuDayCosts[day1][c] = stuDayCost(day1,c,studentavailability);
				stuDayCosts[day2][c] = stuDayCost(day2,c,studentavailability);
				dayCosts[day1] = dayCosts[day1] + (stuDayCosts[day1][c] - d1Orig[c]);
				dayCosts[day2] = dayCosts[day2] + (stuDayCosts[day2][c] - d2Orig[c]);
			}
		}
		else if (ev1 != -1 && ev2 == -1){
			for(i=0; i<eventStus[ev1].size(); i++){
				c = eventStus[ev1][i];
				d1Orig[c] = stuDayCosts[day1][c]; 
				d2Orig[c] = stuDayCosts[day2][c]; 
				studentavailability[slot1][c] = true; 
				studentavailability[slot2][c] = false;
				stuDayCosts[day1][c] = stuDayCost(day1,c,studentavailability);
				stuDayCosts[day2][c] = stuDayCost(day2,c,studentavailability);
				dayCosts[day1] = dayCosts[day1] + (stuDayCosts[day1][c] - d1Orig[c]);
				dayCosts[day2] = dayCosts[day2] + (stuDayCosts[day2][c] - d2Orig[c]);
			}
		}
		else {
			cout << "Error in RecalculateStudentTimetables\n";
			exit(1);
		}
	}
}
inline
void resetStudentTimetables(int ev1, int ev2, int slot1, int slot2, IntVector &d1Orig, IntVector &d2Orig, bool **studentavailability, int **stuDayCosts, int *dayCosts)
{
	//Resets the cost change after a simple move (i.e. a move or a swap)
	int day1 = slot1/SLOTSPERDAY; int day2 = slot2/SLOTSPERDAY, c, diff1, diff2, i;
	if (day1 == day2){
		//We've only changed 1 day. We only need to undo the students involved in this
		if (ev1 != -1 && ev2 != -1){
			for(c=0; c<numStudents; c++){
				if (attends[c][ev1] != attends[c][ev2]){
					diff1 = d1Orig[c]-stuDayCosts[day1][c];
					stuDayCosts[day1][c] = d1Orig[c];
					studentavailability[slot1][c] = !studentavailability[slot1][c];
					studentavailability[slot2][c] = !studentavailability[slot2][c];
					dayCosts[day1] = dayCosts[day1] + diff1;
				}
			}
		}
		else if (ev1 == -1 && ev2 != -1){ 
			//ev1 is a blank and ev2 is not
			for(i=0; i<eventStus[ev2].size(); i++){
				c = eventStus[ev2][i];
				//student c is involved in the swap so we want to change his tt
				diff1 = d1Orig[c]-stuDayCosts[day1][c];
				stuDayCosts[day1][c] = d1Orig[c];
				studentavailability[slot1][c] = false; 
				studentavailability[slot2][c] = true;
				dayCosts[day1]= dayCosts[day1]+diff1;
			}
		}
		else if (ev1 != -1 && ev2 == -1){
			for(i=0; i<eventStus[ev1].size(); i++){
				c = eventStus[ev1][i];
				diff1 = d1Orig[c]-stuDayCosts[day1][c];
				stuDayCosts[day1][c] = d1Orig[c];
				studentavailability[slot1][c] = true; 
				studentavailability[slot2][c] = false;
				dayCosts[day1] = dayCosts[day1] + diff1;
			}
		}
		else {
			cout << "Error in RecalculateStudentTimetables\n";
			exit(1);
		}	
	}
	else {
		//Weve changed two days so need to recalculate them both
		if (ev1 != -1 && ev2 != -1){
			//Swapping two events
			for(c=0; c<numStudents; c++){
				if (attends[c][ev1] != attends[c][ev2]){
					//the student s is taking one of the events. We now alter the Savail and
					//sort the stuDayCosts out
					diff1 = d1Orig[c]-stuDayCosts[day1][c];
					diff2 = d2Orig[c]-stuDayCosts[day2][c];
					stuDayCosts[day1][c] = d1Orig[c];
					stuDayCosts[day2][c] = d2Orig[c];
					studentavailability[slot1][c] = !studentavailability[slot1][c];
					studentavailability[slot2][c] = !studentavailability[slot2][c];
					dayCosts[day1] = dayCosts[day1] + diff1;
					dayCosts[day2] = dayCosts[day2] + diff2;
				}
			}
		}
		else if (ev1 == -1 && ev2 != -1){
			//ev1 is a blank and ev2 is not
			for(i=0; i<eventStus[ev2].size(); i++){
				c = eventStus[ev2][i];
				//student c is involved in the swap so we want to change his tt
				diff1 = d1Orig[c]-stuDayCosts[day1][c];
				diff2 = d2Orig[c]-stuDayCosts[day2][c];
				stuDayCosts[day1][c] = d1Orig[c];
				stuDayCosts[day2][c] = d2Orig[c]; 
				studentavailability[slot1][c] = false; 
				studentavailability[slot2][c] = true;
				dayCosts[day1] = dayCosts[day1] + diff1;
				dayCosts[day2] = dayCosts[day2] + diff2;
			}
		}
		else if (ev1 != -1 && ev2 == -1){
			for(i=0; i<eventStus[ev1].size(); i++){
				c = eventStus[ev1][i];
				diff1 = d1Orig[c]-stuDayCosts[day1][c];
				diff2 = d2Orig[c]-stuDayCosts[day2][c];
				stuDayCosts[day1][c] = d1Orig[c];
				stuDayCosts[day2][c] = d2Orig[c];
				studentavailability[slot1][c] = true; 
				studentavailability[slot2][c] = false;
				dayCosts[day1] = dayCosts[day1] + diff1;
				dayCosts[day2] = dayCosts[day2] + diff2;
			}
		}
		else {
			cout << "Error in RecalculateStudentTimetables\n";
			exit(1);
		}
	}
}
//-----------------------------------------------------------------------------------
inline
bool equalStatus(TwoDIntVector &theCols, TwoDIntVector &theGroups, int slot1, int slot2, int s)
{
	//Returns true iff: (a) student s attends two events involved in the interchange (i.e. one in each slot)
	//					(b) student s attends no events involved in the interchange 
	bool at1 = false, at2 = false;
	int i;
	for(i=0;i<theCols[0].size();i++){
		if(theCols[0][i] == BLACK && attends[s][theGroups[0][i]]){
			at1 = true;
			break;
		}
	}
	for(i=0;i<theCols[1].size();i++){
		if(theCols[1][i] == BLACK && attends[s][theGroups[1][i]]){
			at2 = true;
			break;
		}
	}
	if(at1 == at2) return true;
	else return false;
}
inline
void recalculateStudentTimetablesKempe(int slot1, int slot2, IntVector &d1Orig, IntVector &d2Orig, bool **studentavailability, int **stuDayCosts, int *dayCosts, 
				TwoDIntVector &theCols, TwoDIntVector &theGroups)
{
	//Calculates the cost change after a Kempe chain interchange
	int day1 = slot1/SLOTSPERDAY; int day2 = slot2/SLOTSPERDAY, c;
	if (day1 == day2){
		//We've only changed 1 day. We only need to recalculate the students involved in this
		for(c=0; c<numStudents; c++){
			if(!equalStatus(theCols,theGroups,slot1,slot2,c)){
				//the student s is taking one of the events. We now alter the Savail and sort the stuDayCosts out
				d1Orig[c] = stuDayCosts[day1][c]; 
				studentavailability[slot1][c] = !studentavailability[slot1][c];
				studentavailability[slot2][c] = !studentavailability[slot2][c];
				stuDayCosts[day1][c] = stuDayCost(day1,c,studentavailability);
				dayCosts[day1] = dayCosts[day1] + (stuDayCosts[day1][c] - d1Orig[c]);
			}
		}
	}
	else {
		//Weve changed two days so need to recalculate them both
		for(c=0; c<numStudents; c++){
			if(!equalStatus(theCols,theGroups,slot1,slot2,c)){
				//the student s is taking one of the events. We now alter the Savail and sort the stuDayCosts out
				d1Orig[c] = stuDayCosts[day1][c]; 
				d2Orig[c] = stuDayCosts[day2][c]; 
				studentavailability[slot1][c] = !studentavailability[slot1][c];
				studentavailability[slot2][c] = !studentavailability[slot2][c];
				stuDayCosts[day1][c] = stuDayCost(day1,c,studentavailability);
				stuDayCosts[day2][c] = stuDayCost(day2,c,studentavailability);
				dayCosts[day1] = dayCosts[day1] + (stuDayCosts[day1][c] - d1Orig[c]);
				dayCosts[day2] = dayCosts[day2] + (stuDayCosts[day2][c] - d2Orig[c]);
			}
		}
	}
}
inline
void resetStudentTimetablesKempe(int slot1, int slot2, IntVector &d1Orig, IntVector &d2Orig, bool **studentavailability, int **stuDayCosts, int *dayCosts, 
				TwoDIntVector &theCols, TwoDIntVector &theGroups)
{
	//Resets the cost change after a Kempe chain interchange
	int day1 = slot1/SLOTSPERDAY; int day2 = slot2/SLOTSPERDAY, c, diff1, diff2;
	if (day1 == day2){
		//Weve only changed 1 day. We only need to undo the students involved in this
		for(c=0; c<numStudents; c++){
			if(!equalStatus(theCols,theGroups,slot1,slot2,c)){
				diff1 = d1Orig[c]-stuDayCosts[day1][c];
				stuDayCosts[day1][c] = d1Orig[c];
				studentavailability[slot1][c] = !studentavailability[slot1][c];
				studentavailability[slot2][c] = !studentavailability[slot2][c];
				dayCosts[day1] = dayCosts[day1] + diff1;
			}
		}
	}
	else {
		//Weve changed two days so need to reset them both
		for(c=0; c<numStudents; c++){
			if(!equalStatus(theCols,theGroups,slot1,slot2,c)){
				//the student s is taking one of the events. We now alter the Savail and
				//sort the stuDayCosts out
				diff1 = d1Orig[c]-stuDayCosts[day1][c];
				diff2 = d2Orig[c]-stuDayCosts[day2][c];
				stuDayCosts[day1][c] = d1Orig[c];
				stuDayCosts[day2][c] = d2Orig[c];
				studentavailability[slot1][c] = !studentavailability[slot1][c];
				studentavailability[slot2][c] = !studentavailability[slot2][c];
				dayCosts[day1] = dayCosts[day1] + diff1;
				dayCosts[day2] = dayCosts[day2] + diff2;
			}
		}
	}
}
//-----------------------------------------------------------------------------------
inline
bool thereAreNoPrecViolationsSC(IntVector &slotOfEvent, int ev1, int ev2)
{	
	//Do a quick check to see if the events involved in the move/swap are not associated with any prec constr.
	if(!hasPrecConstr[ev1] && ev2 == -1) return true;
	else if(!hasPrecConstr[ev1] && !hasPrecConstr[ev2]) return true;

	int i;
	for(i=0; i<precidence[ev1].size(); i++){
		//Check timeslot of all events that have to come after ev1
		if(slotOfEvent[precidence[ev1][i]] != -1) {
			if(slotOfEvent[precidence[ev1][i]] <= slotOfEvent[ev1]) return false; 
		}
	}
	for(i=0; i<beforePrec[ev1].size(); i++){
		//Check timeslot of all events that have to come before ev1
		if(slotOfEvent[beforePrec[ev1][i]] != -1) {
			if(slotOfEvent[beforePrec[ev1][i]] >= slotOfEvent[ev1]) return false; 
		}
	}

	if(ev2 != -1){
		for(i=0; i<precidence[ev2].size(); i++){
			//Check timeslot of all events that have to come after ev2
			if(slotOfEvent[precidence[ev2][i]] != -1) {
				if(slotOfEvent[precidence[ev2][i]] <= slotOfEvent[ev2]) return false; 
			}
		}
		for(i=0; i<beforePrec[ev2].size(); i++){
			//Check timeslot of all events that have to come before ev2
			if(slotOfEvent[beforePrec[ev2][i]] != -1) {
				if(slotOfEvent[beforePrec[ev2][i]] >= slotOfEvent[ev2]) return false; 
			}
		}
	}
	return true;
}
inline
bool thereAreNoPrecViolationsSC(IntVector &slotOfEvent, int ev)
{	
	//Overloaded version. Check precidence violations with just the one event
	if(!hasPrecConstr[ev]) return true;
	int i;
	for(i=0; i<precidence[ev].size(); i++){
		//Check timeslot of all events that have to come after ev2
		if(slotOfEvent[precidence[ev][i]] != -1) {
			if(slotOfEvent[precidence[ev][i]] <= slotOfEvent[ev]) return false; 
		}
	}
	for(i=0; i<beforePrec[ev].size(); i++){
		//Check timeslot of all events that have to come before ev2
		if(slotOfEvent[beforePrec[ev][i]] != -1) {
			if(slotOfEvent[beforePrec[ev][i]] >= slotOfEvent[ev]) return false; 
		}
	}
	return true;
}

//-----------------------------------------------------------------------------------
inline
void pickAtomsSC(int &r1, int &c1, int &r2, int &c2, bool &doingSwap, int &ev1, int &ev2, TTAndUnplaced &theSolution)
{				
	do{
		r1 = rand()%numRooms; 
		c1 = rand()%NUMBEROFSLOTS;
	} while(theSolution.TT[r1][c1] == -1);
	
	//Now choose any cell in a different timeslot (can be blank or non blank)
	do{
		r2 = rand()%numRooms; 
		c2 = rand()%NUMBEROFSLOTS;
	} while(c2 == c1);

	if(theSolution.TT[r2][c2]== -1)
		doingSwap = false;
	else 
		doingSwap = true;

	ev1 = theSolution.TT[r1][c1];
	ev2 = theSolution.TT[r2][c2];
}

//-----------------------------------------------------------------------------------
inline 
void resetMoveSC(int c1, int r1, int c2, int r2, IntVector &c1Orig, IntVector &c2Orig, IntVector &timeSlotForEvent, bool doingSwap, TTAndUnplaced &theSolution)
{
	//Resets a move (move, swap, Kempe) by reverting to the copied slots
	int r;
	for(r=0;r<numRooms;r++){
		theSolution.TT[r][c1] = c1Orig[r];
		theSolution.TT[r][c2] = c2Orig[r];
		if(c1Orig[r] != -1) timeSlotForEvent[c1Orig[r]] = c1;
		if(c2Orig[r] != -1) timeSlotForEvent[c2Orig[r]] = c2;
	}
}


//-----------------------------------------------------------------------------------
inline 
bool maxMatchSC(int evToInsert, int j, IntVector &col, IntVector &pred, IntQueue &Q, TTAndUnplaced &theSolution)
{
	//Maximum matching algorithm based on Bredth first search. Given "feasible" timeslot j, we have an event to insert
	//that is known not to clash with any events in the timeslot, but is not suitable for any of the available rooms.
	//Thus the procedure below is run. If it is succesful, then the changes are made and done = true; 
	//otherwise, no changes are made and done is false.

	int endPoint,r,ev,rm,v,u;	

	//initialise data structures needed for BFS
	col.clear();
	col.resize(numRooms*2+1, WHITE);
	pred.clear();
	pred.resize(numRooms*2+1, NIL);
	Q.clear();
	col.back() = GREY;
	Q.push_back(numRooms*2);
	
	while(!Q.empty()){
		u = Q.front();
		if(u>=numRooms){
			//We are currently at an "event node" u. Check to see if there is a room available for it
			//First determine the event "ev" we are dealing with
			if(u==numRooms*2) 
				ev = evToInsert;
			else 
				ev = theSolution.TT[u-numRooms][j];
			for(r=0;r<feasRooms[ev].size();r++){
				rm = feasRooms[ev][r];
				if(theSolution.TT[rm][j]==-1){
					//Room "rm" is suitable and available for event u so we end
					pred[rm] = u;
					endPoint = rm;
					goto DONE;
				}
				if(col[rm]==WHITE){
					//Room "rm" is suitable, but not available, so we will explore from it
					col[rm]=GREY;
					pred[rm] = u;
					Q.push_back(rm);
				}
			}
		}
		else {
			//We are currently at a room node u that is occupied, so we traverse to the event node that is occupying it
			v = u+numRooms;
			if(col[v]==WHITE){
				col[v]=GREY;
				pred[v] = u;
				Q.push_back(v);
			}
		}
		Q.pop_front();
		col[u] = BLACK;
	}
	//If we are here then a matching is not possible.
	return false;

DONE:
	//If we are here then a matching was found. Thus we now make the changes to the timetable
	int current = endPoint;
	while(pred[pred[current]] != NIL){
		theSolution.TT[current][j] = theSolution.TT[pred[current]-numRooms][j];
		current = pred[pred[current]];
	}
	theSolution.TT[current][j] = evToInsert;
	return true;	
}
//-------------------------------------------------------------------------------
inline
bool trySimpleMove(int r1, int c1, int r2, int c2, IntVector &timeSlotForEvent,
			   IntVector &col, IntVector &pred, IntQueue &Q, TTAndUnplaced &theSolution,
			   IntVector &c1Orig, IntVector &c2Orig, bool &changed)
{
	//If the change can be made, we do so, otherwise we simply exit (no need to reset -- this isdone in the calling function
	int r, ev1 = theSolution.TT[r1][c1], ev2 = theSolution.TT[r2][c2];
	
	//If this is set to true, it means that timeslots c1 and c2 have changed (and copies have been made, which will need to be used if reset)
	changed = false;

	//MOVE EVENT ev1 to c2
	if(ev2 == -1){
		
		//We only want to see if we can put ev1 into slot c2. First check slot would be feasible  
		if(!eventAvail[ev1][c2]) 
			return false;
		for(r=0;r<numRooms;r++)	
			if(theSolution.TT[r][c2] != -1 && conflict[theSolution.TT[r][c2]][ev1]) 
				return false;
		
		//We now might have to make changes to c1 and c2 so we make a copies
		changed = true;
		for(r=0;r<numRooms;r++){
			c1Orig[r] = theSolution.TT[r][c1]; 
			c2Orig[r] = theSolution.TT[r][c2];
		}

		//Now try to insert the ev1 into any available blanks in c2
		for(r=0;r<feasRooms[ev1].size();r++){
			if(theSolution.TT[feasRooms[ev1][r]][c2]==-1){
				//Room is feasible for ev1 so we make the changes and then check for prec violations
				theSolution.TT[feasRooms[ev1][r]][c2] = ev1;
				theSolution.TT[r1][c1] = -1;
				timeSlotForEvent[ev1] = c2;
				if(thereAreNoPrecViolationsSC(timeSlotForEvent,ev1,ev2))
					return true;
				else
					return false;
			}
		}
		//We now try to insert ev1 into c2 using max Matching algorithm (if felxibility > 1)
		if(flexibility > 1 && maxMatchSC(ev1,c2,col,pred,Q,theSolution)){
			theSolution.TT[r1][c1] = -1;
			timeSlotForEvent[ev1] = c2;
			if(thereAreNoPrecViolationsSC(timeSlotForEvent,ev1,ev2))
				return true;
			else
				return false;
		}
		else
			//If we are here then no change was made and so we return the original settings
			return false;
	}
	//Swap events ev1 and ev2
	else {
		
		//We want to see if we can put ev1 into slot c2 and ev2 into slot c1  
		if(!eventAvail[ev1][c2] || !eventAvail[ev2][c1]) 
			return false;

		//Remove the events and check for clashes. If there are any we reset and exit immediately
		theSolution.TT[r1][c1] = -1; 
		theSolution.TT[r2][c2] = -1;
		for(r=0;r<numRooms;r++) {
			if(theSolution.TT[r][c2] != -1 && conflict[theSolution.TT[r][c2]][ev1]){
				theSolution.TT[r1][c1] = ev1; 
				theSolution.TT[r2][c2] = ev2;
				return false;
			}
			if(theSolution.TT[r][c1] != -1 && conflict[theSolution.TT[r][c1]][ev2]){
				theSolution.TT[r1][c1] = ev1; 
				theSolution.TT[r2][c2] = ev2;
				return false;
			}
		}
		
		//If we are here we might have to make changes to c1 and c2 so we make a copies
		changed = true;
		for(r=0;r<numRooms;r++){
			c1Orig[r] = theSolution.TT[r][c1]; 
			c2Orig[r] = theSolution.TT[r][c2];
		}
		//(don't forget to include ev1 and ev2, which were deleted from TT above).
		c1Orig[r1]=ev1;
		c2Orig[r2]=ev2;
		
		//We now want to check for room availability. 
		bool doneEv1Move = false, doneEv2Move = false;

		//First try to insert ev1 into c2
		for(r=0;r<feasRooms[ev1].size();r++){
			if(theSolution.TT[feasRooms[ev1][r]][c2]==-1){
				//Room is feasible for ev1 so add it in to the timetable
				theSolution.TT[feasRooms[ev1][r]][c2] = ev1;
				doneEv1Move = true;
				break;
			}
		}
		if(!doneEv1Move){
			//We now try to insert ev1 into c2 using max Matching algorithm
			if(flexibility > 1)
				doneEv1Move = maxMatchSC(ev1,c2,col,pred,Q,theSolution);
		}
		if(!doneEv1Move){
			//The first swap could not be made so we exit
			return false;
		} 
		else {
			//Ev1 has been moved to slot c2. So we also try the same thing, putting ev2 into c1
			for(r=0;r<feasRooms[ev2].size();r++){
				if(theSolution.TT[feasRooms[ev2][r]][c1]==-1){
					//Room is feasible for ev2 so we make the changes. The whole swap is now feasible
					theSolution.TT[feasRooms[ev2][r]][c1] = ev2;
					timeSlotForEvent[ev1] = c2;
					timeSlotForEvent[ev2] = c1;
					if(thereAreNoPrecViolationsSC(timeSlotForEvent,ev1,ev2))
						return true;
					else
						return false;
				}
			}
			//If we are here we now want to try to insert ev2 into c1 using max Matching algorithm
			if(flexibility > 1)
				doneEv2Move = maxMatchSC(ev2,c1,col,pred,Q,theSolution);
			if(doneEv2Move){
				timeSlotForEvent[ev1] = c2;
				timeSlotForEvent[ev2] = c1;
				if(thereAreNoPrecViolationsSC(timeSlotForEvent,ev1,ev2))
					return true;
				else
					return false;
			}
			else {
				//We have tried everything. The swap cannot be performed
				return false;
			}
		}
	}
}
//-------------------------------------------------------------------------------------
inline
void DFSVisit(int uPos, int theGroup, TwoDIntVector &theCols, TwoDIntVector &theGroups, int &blackCnt)
{
	int theOtherGroup=0, vPos;
	if(theGroup == 0) theOtherGroup = 1;
	theCols[theGroup][uPos] = GREY;
	for(vPos = 0; vPos<theCols[theOtherGroup].size(); vPos++){
		if(conflict[theGroups[theGroup][uPos]][theGroups[theOtherGroup][vPos]]){
			if(theCols[theOtherGroup][vPos] == WHITE){
				DFSVisit(vPos,theOtherGroup,theCols,theGroups,blackCnt);
			}
		}
	}
	theCols[theGroup][uPos] = BLACK;
	blackCnt++;
}
inline
int IDKempeChain(TTAndUnplaced &sol, int slot1, int r1, int slot2, TwoDIntVector &theCols, TwoDIntVector &theGroups){
	int r, blackCnt = 0, nodePos;
	//Set up the vectors that will hold the result of DFS (and therefore tell us what events are in our Kempe chain)
	theCols[0].clear();
	theCols[1].clear();
	theGroups[0].clear();
	theGroups[1].clear();
	for(r=0; r<numRooms; r++) {
		if(sol.TT[r][slot1] != -1){
			theCols[0].push_back(WHITE);
			theGroups[0].push_back(sol.TT[r][slot1]);
			if(r==r1) nodePos = theCols[0].size()-1;
		}
		if(sol.TT[r][slot2] != -1){
			theCols[1].push_back(WHITE);
			theGroups[1].push_back(sol.TT[r][slot2]);
		}
	}
	//Now start the restricted DFS from the chosen node
	DFSVisit(nodePos,0,theCols,theGroups,blackCnt);
	//We can now ID the nodes that are in the Kempe chain by seeing which ones are BLACK.
	return(blackCnt);
}
//-------------------------------------------------------------------------------------
inline
bool tryKempeMove(IntVector &slotOfEvent, int slot1, int slot2, IntVector &col, IntVector &pred, IntQueue &Q, TTAndUnplaced &theSolution, IntVector &c1Orig, IntVector &c2Orig, 
				bool &changed, TwoDIntVector &theCols, TwoDIntVector &theGroups)
{
	int i, r, s1Black=0, s1White=0, s2Black=0, s2White=0, ev, numKempeChains=1;
	bool doneMove, slotsWillBeOverFull = false, swappingFullSlots = false;
	//If this is set to true, it means that timeslots slot1 and slot2 have changed (and copies have been made, which will need to be used if reset)
	changed = false;
	//First check that the two slots are suitable for all nodes involved in the interchange
	for(i=0;i<theGroups[0].size();i++){
		if(theCols[0][i] == BLACK){
			s1Black++;
			if(!eventAvail[theGroups[0][i]][slot2]) 
				return false; //The interchange will violate an event avail constraint so end
		} else {
			s1White++;
		}
	}
	for(i=0;i<theGroups[1].size();i++){
		if(theCols[1][i] == BLACK){
			s2Black++;
			if(!eventAvail[theGroups[1][i]][slot1]) 
				return false; //The interchange will violate an event avail constraint so end
		} else {
			s2White++;
		}
	}

	//Check that the interchange won't overfill a slot. if so, we either exit or consider doing another chain
	if(s1White + s2Black > numRooms || s2White + s1Black > numRooms){
		if(numKempeChains >= maxNumKempeChains) return false;
		else slotsWillBeOverFull = true; //We don't do a "goto DOUBLEKEMPE" here because prec violations in the single Kempe will still hold in the double Kempe, so there's no point in calculating the other Kempe until we know if this is the case
	}

	//Check to see if the Kempe is actually a slotSwap. If so, there's no need to run the max_match algorithm if the swap turns out to be OK
	if(s1White == 0 && s2White == 0) swappingFullSlots = true;

	//No need to check for conflicts violations (because it is a Kempe chain), so we next check for precidence violations
	bool precViolation = false;
	//Make the relevant changes to the slotOfEvent vector
	for(i=0;i<theGroups[0].size();i++)	if(theCols[0][i] == BLACK)	slotOfEvent[theGroups[0][i]] = slot2;
	for(i=0;i<theGroups[1].size();i++)	if(theCols[1][i] == BLACK)	slotOfEvent[theGroups[1][i]] = slot1;
	//Now check for prec constraints with the relevant events
	for(i=0;i<theGroups[0].size();i++){
		if(theCols[0][i] == BLACK && !thereAreNoPrecViolationsSC(slotOfEvent,theGroups[0][i])){
			precViolation = true;
			break;
		}
	}
	if(!precViolation){
		for(i=0;i<theGroups[1].size();i++){
			if(theCols[1][i] == BLACK && !thereAreNoPrecViolationsSC(slotOfEvent,theGroups[1][i])){
				precViolation = true;
				break;
			}
		}
	}
	//If there is a precidence violation then we reset to the original and end
	if(precViolation){
		for(i=0;i<theGroups[0].size();i++) slotOfEvent[theGroups[0][i]] = slot1;
		for(i=0;i<theGroups[1].size();i++) slotOfEvent[theGroups[1][i]] = slot2;
		return false;
	}

	//If we are here then we will attempt to make the change so first copy the original slots (remember that slotoOfEvents has already been altered)
	changed = true;
	for(r=0;r<numRooms;r++){
		c1Orig[r] = theSolution.TT[r][slot1];
		c2Orig[r] = theSolution.TT[r][slot2];
	}

	if(slotsWillBeOverFull) goto MULTIPLEKEMPE;

	if(swappingFullSlots){
		//No need to do maxMatch etc. because ID Kempe chain move is equivalent to a swapSlots operation
		for(r=0;r<numRooms;r++)	swapInts(theSolution.TT[r][slot1],theSolution.TT[r][slot2]);
		NumKempeChainsRecord[numKempeChains]++;
		return true;
	}

	//First, remove the events involved in the interchange from slot1
	r = 0;
	for(i=0;i<theGroups[0].size();i++){
		if(theCols[0][i] == BLACK){
			while(true){
				if(theSolution.TT[r][slot1] == theGroups[0][i]) break;
				else r++;
			}
			theSolution.TT[r][slot1] = -1;
		}
	}
	//Now, attempt to insert all events involved in the interchange from slot2 into rooms in slot1
	for(i=0;i<theGroups[1].size();i++){
		if(theCols[1][i] == BLACK){
			doneMove = false;
			ev = theGroups[1][i];
			for(r=0;r<feasRooms[ev].size();r++){
				if(theSolution.TT[feasRooms[ev][r]][slot1]==-1){
					//Room is feasible for ev so add it in to the timetable
					theSolution.TT[feasRooms[ev][r]][slot1] = ev;
					doneMove = true;
					break;
				}
			}
			//If needed, we now try to insert ev into slot1 using max Matching algorithm
			if(!doneMove) doneMove = maxMatchSC(ev,slot1,col,pred,Q,theSolution);
			//If we've still had no success then ev cannot be inserted into slot1 so we exit or try a double kempe chain
			if(!doneMove){
				if(numKempeChains >= maxNumKempeChains) return false;
				else goto MULTIPLEKEMPE;
			}
		}
	}
	//Now remove the events involved in the interchange from slot2
	r = 0;
	for(i=0;i<theGroups[1].size();i++){
		if(theCols[1][i] == BLACK){
			while(true){
				if(theSolution.TT[r][slot2] == theGroups[1][i]) break;
				else r++;
			}
			theSolution.TT[r][slot2] = -1;
		}
	}
	//And attempt to insert all events involved in the interchange from slot1 into rooms in slot2
	for(i=0;i<theGroups[0].size();i++){
		if(theCols[0][i] == BLACK){
			doneMove = false;
			ev = theGroups[0][i];
			for(r=0;r<feasRooms[ev].size();r++){
				if(theSolution.TT[feasRooms[ev][r]][slot2]==-1){
					//Room is feasible for ev so add it in to the timetable
					theSolution.TT[feasRooms[ev][r]][slot2] = ev;
					doneMove = true;
					break;
				}
			}
			//If needed we now try to insert ev into slot2 using max Matching algorithm
			if(!doneMove) doneMove = maxMatchSC(ev,slot2,col,pred,Q,theSolution);
			//If we've still had no success then ev cannot be inserted into slot2 so we exit
			if(!doneMove){
				if(numKempeChains >= maxNumKempeChains) return false;
				else goto MULTIPLEKEMPE;
			}
		}
	}
	//If we are here then the single Kempe interchange will maintain feasibility
	NumKempeChainsRecord[numKempeChains]++;
	return true;

MULTIPLEKEMPE:
	numKempeChains++;
	//We will only be here if flexibility is set to >=4 and (a) The proposed Kempe move(s) result in too many events in a slot, or 
	//(b) we have been unable to identify a room for an event in the proposed Kempe interchange.
	//So we now attempt a double Kempe interchange. First reset any changes made above to the solution
	for(r=0;r<numRooms;r++){
		theSolution.TT[r][slot1] = c1Orig[r];
		theSolution.TT[r][slot2] = c2Orig[r];
		if(c1Orig[r] != -1) slotOfEvent[c1Orig[r]] = slot1;
		if(c2Orig[r] != -1) slotOfEvent[c2Orig[r]] = slot2;
	}
	//Now ID a random white node (i.e. one that isn't in the first Kempe chain)
	int numChoices = 0, nodePos=-1, theGroup, cSize;
	for(i=0;i<theCols[0].size();i++){
		if(theCols[0][i]==WHITE){
			if(rand()%(numChoices+1)==0){
				nodePos = i;
				theGroup = 0;
			}
			numChoices++;
		}
	}
	for(i=0;i<theCols[1].size();i++){
		if(theCols[1][i]==WHITE){
			if(rand()%(numChoices+1)==0){
				nodePos = i;
				theGroup = 1;
			}
			numChoices++;
		}
	}
	//Now form a new, Kempe chain from this white node (this will have no common nodes/events with previous Kempe chains)
	if(nodePos==-1){cout<<"Error: should always be a white node present at this stage\n"; exit(1);}
	DFSVisit(nodePos,theGroup,theCols,theGroups,cSize);

	//We now attempt the multiple move in the same way as the single move (by switching the black ones)
	s1Black=0; s1White=0; s2Black=0; s2White=0;
	slotsWillBeOverFull = false; swappingFullSlots = false; changed = false;
	//First check that the two slots are suitable for all nodes involved in the multiple Kempe interchange
	for(i=0;i<theGroups[0].size();i++){
		if(theCols[0][i] == BLACK){
			s1Black++;
			if(!eventAvail[theGroups[0][i]][slot2]) return false; //The multiple interchange will violate an event avail constraint so end
		} 
		else s1White++;
	}
	for(i=0;i<theGroups[1].size();i++){
		if(theCols[1][i] == BLACK){
			s2Black++;
			if(!eventAvail[theGroups[1][i]][slot1]) return false; //The multiple interchange will violate an event avail constraint so end
		} 
		else s2White++;
	}
	//Check that the multiple interchange won't overfill a slot. If it does we either do another Kempe or exit
	if(s1White + s2Black > numRooms || s2White + s1Black > numRooms){
		if(numKempeChains >= maxNumKempeChains) return false;
		else slotsWillBeOverFull = true; 
	}

	//Check to see if the double interchange is actually a slotSwap. If so, there's no need to run the max_match algorithm
	//if the swap turns out to be OK
	if(s1White == 0 && s2White == 0) swappingFullSlots = true;

	precViolation = false;
	//Make the relevant changes to the slotOfEvent vector
	for(i=0;i<theGroups[0].size();i++)	if(theCols[0][i] == BLACK)	slotOfEvent[theGroups[0][i]] = slot2;
	for(i=0;i<theGroups[1].size();i++)	if(theCols[1][i] == BLACK)	slotOfEvent[theGroups[1][i]] = slot1;
	//Now check for prec constraints with the relevant events
	for(i=0;i<theGroups[0].size();i++){
		if(theCols[0][i] == BLACK && !thereAreNoPrecViolationsSC(slotOfEvent,theGroups[0][i])){
			precViolation = true;
			break;
		}
	}
	if(!precViolation){
		for(i=0;i<theGroups[1].size();i++){
			if(theCols[1][i] == BLACK && !thereAreNoPrecViolationsSC(slotOfEvent,theGroups[1][i])){
				precViolation = true;
				break;
			}
		}
	}
	//If there is a precidence violation then we reset to the original and end
	if(precViolation){
		for(i=0;i<theGroups[0].size();i++) slotOfEvent[theGroups[0][i]] = slot1;
		for(i=0;i<theGroups[1].size();i++) slotOfEvent[theGroups[1][i]] = slot2;
		return false;
	}

	//Our multiple Kempe operator still results in overfull slots, so see if we can equal this out via a further one
	if(slotsWillBeOverFull) goto MULTIPLEKEMPE;
	
	//If we are here then we will attempt to make the multiple interchange (changing the TT). Remember that slotOfEvents has already been altered
	changed = true;
	if(swappingFullSlots){
		//No need to do maxMatch etc. because the multiple Kempe chain move is equivalent to a swapSlots operation
		for(r=0;r<numRooms;r++)	swapInts(theSolution.TT[r][slot1],theSolution.TT[r][slot2]);
		NumKempeChainsRecord[numKempeChains]++;
		return true;
	}

	//First, remove the events involved in the interchange from slot1
	r = 0;
	for(i=0;i<theGroups[0].size();i++){
		if(theCols[0][i] == BLACK){
			while(true){
				if(theSolution.TT[r][slot1] == theGroups[0][i]) break;
				else r++;
			}
			theSolution.TT[r][slot1] = -1;
		}
	}
	//Now, attempt to insert all events involved in the interchange from slot2 into rooms in slot1
	for(i=0;i<theGroups[1].size();i++){
		if(theCols[1][i] == BLACK){
			doneMove = false;
			ev = theGroups[1][i];
			for(r=0;r<feasRooms[ev].size();r++){
				if(theSolution.TT[feasRooms[ev][r]][slot1]==-1){
					//Room is feasible for ev so add it in to the timetable
					theSolution.TT[feasRooms[ev][r]][slot1] = ev;
					doneMove = true;
					break;
				}
			}
			//If needed, we now try to insert ev into slot1 using max Matching algorithm
			if(!doneMove) doneMove = maxMatchSC(ev,slot1,col,pred,Q,theSolution);
			//If we've still had no success then ev cannot be inserted into slot1 so we exit
			if(!doneMove){
				if(numKempeChains >= maxNumKempeChains) return false;
				else goto MULTIPLEKEMPE;
			}
		}
	}
	//Now remove the events involved in the interchange from slot2
	r = 0;
	for(i=0;i<theGroups[1].size();i++){
		if(theCols[1][i] == BLACK){
			while(true){
				if(theSolution.TT[r][slot2] == theGroups[1][i]) break;
				else r++;
			}
			theSolution.TT[r][slot2] = -1;
		}
	}
	//And attempt to insert all events involved in the interchange from slot1 into rooms in slot2
	for(i=0;i<theGroups[0].size();i++){
		if(theCols[0][i] == BLACK){
			doneMove = false;
			ev = theGroups[0][i];
			for(r=0;r<feasRooms[ev].size();r++){
				if(theSolution.TT[feasRooms[ev][r]][slot2]==-1){
					//Room is feasible for ev so add it in to the timetable
					theSolution.TT[feasRooms[ev][r]][slot2] = ev;
					doneMove = true;
					break;
				}
			}
			//If needed we now try to insert ev into slot2 using max Matching algorithm
			if(!doneMove) doneMove = maxMatchSC(ev,slot2,col,pred,Q,theSolution);
			//If we've still had no success then ev cannot be inserted into slot2 so we exit
			if(!doneMove){
				if(numKempeChains >= maxNumKempeChains) return false;
				else goto MULTIPLEKEMPE;
			}
		}
	}
	//If we are here then the multiple Kempe interchange will maintain feasibility
	NumKempeChainsRecord[numKempeChains]++;
	return true;
}
	
//-------------------------------------------------------------------------------------
inline 
bool moveFeasible(int r1, int c1, int r2, int c2, IntVector &timeSlotForEvent,
			   IntVector &col, IntVector &pred, IntQueue &Q, TTAndUnplaced &theSolution,
			   IntVector &c1Orig, IntVector &c2Orig, bool &changed, TwoDIntVector &theCols, TwoDIntVector &theGroups, bool &doneKempe)
{
	doneKempe = false;
	if(flexibility <= 2){
		//We attempt a simple "move" or "swap" operation
		return trySimpleMove(r1,c1,r2,c2,timeSlotForEvent,col,pred,Q,theSolution,c1Orig,c2Orig,changed);
	}
	
	//If we are here we will attempt a move from the the union of Kempe and swap operators.
	//First, check if we are just attempting a swap of 2 con-clashing events. If so, try the swap and end
	if(theSolution.TT[r1][c1] != -1 && theSolution.TT[r2][c2] != -1 && !conflict[theSolution.TT[r1][c1]][theSolution.TT[r2][c2]]){
		return trySimpleMove(r1,c1,r2,c2,timeSlotForEvent,col,pred,Q,theSolution,c1Orig,c2Orig,changed);
	}
		
	//If we are here, we need to ID a Kempe chain starting from ev1 and involving slot c2
	int chainSize = IDKempeChain(theSolution,c1,r1,c2,theCols,theGroups);
	if(chainSize <= 2){
		//The ID'd Kempe chain is a simple move or swap operation so do it and end
		return trySimpleMove(r1,c1,r2,c2,timeSlotForEvent,col,pred,Q,theSolution,c1Orig,c2Orig,changed);
	} else {
		//If we are here The ID'd Kempe chain will involve a move outside of the simple swap/move operation, so we need to check feasibility
		doneKempe = true;
		return tryKempeMove(timeSlotForEvent,c1,c2,col,pred,Q,theSolution,c1Orig,c2Orig,changed,theCols,theGroups);	
	}
}

//-----------------------------------------------------------------------------------
inline
void determineInitTemp(double &initTemp, clock_t &totalTimeForSASoft, IntVector &timeSlotForEvent,
					 int &softCost, IntVector &col, IntVector &pred, IntQueue &Q, TTAndUnplaced &theSolution, 
					 bool **studentavailability, int **stuDayCosts, int *dayCosts, TwoDIntVector &theCols, TwoDIntVector &theGroups,
					 int &dummyCost, int &evsInDummy)
{
	int sample[100], sampleSize = 1, r1, c1, r2, c2, ev1, ev2,proposedDummyCost,proposedEvsInDummy;
	sample[0] = softCost+dummyCost;
	IntVector c1Orig(numRooms),c2Orig(numRooms),d1Orig(numStudents),d2Orig(numStudents);
	bool doingSwap, changed, doneKempe;
	
	clock_t timeForInitTemp = clock_t(totalTimeForSASoft * 0.01);
	timeForInitTemp = clock() + timeForInitTemp;

	//Work out the initial temperature. This is done for a sample of 100 or until 1% of time is reached
	while(timeForInitTemp > clock() && sampleSize < 100){
		pickAtomsSC(r1,c1,r2,c2,doingSwap,ev1,ev2,theSolution);
		if(moveFeasible(r1,c1,r2,c2,timeSlotForEvent,col,pred,Q,theSolution,c1Orig,c2Orig,changed,theCols,theGroups,doneKempe)){
			if(doneKempe)
				recalculateStudentTimetablesKempe(c1,c2,d1Orig,d2Orig,studentavailability,stuDayCosts,dayCosts,theCols,theGroups);
			else
				recalculateStudentTimetables(ev1,ev2,c1,c2,d1Orig,d2Orig,studentavailability,stuDayCosts,dayCosts);
			softCost = totalCost(dayCosts,NUMDAYS);
			calcNewDummyCost(theSolution,c1,c2,c1Orig,c2Orig,proposedDummyCost,proposedEvsInDummy,dummyCost,evsInDummy);
			dummyCost = proposedDummyCost;
			evsInDummy = proposedEvsInDummy;
			sample[sampleSize] = (weightCoefficient*dummyCost)+softCost;
			sampleSize++;
		}
		else {
			if(changed) resetMoveSC(c1,r1,c2,r2,c1Orig,c2Orig,timeSlotForEvent,doingSwap,theSolution);
		}
	}
	if(sampleSize <= 1){
		cout<<"Error: Problem instance is too constrained for movements in the Feasible-only search space to be performed\n"; 
		exit(1);
	} else {
		initTemp = stDevP(sample, sampleSize);
	}
}

//-------------------------------------------------------------------------------
void doSAPhase2(int &softCost, int &dToF, clock_t &Part3ClockFinish, TTAndUnplaced &theSolution, 
				bool **studentavailability, int **stuDayCosts, int *dayCosts, double SAEndTemp, int &chainCnt)
{
	double alpha, initTemp, temperature, MCTime, timeRemaining, MCsRemaining;
	int r, c, r1, r2, c1, c2, ev1, ev2, proposedSoftCost, costChange, cnt, mCLen = numEvents*numEvents, recRow=0, dummyCost=0, evsInDummy=0, proposedDummyCost, bestDummyCost=0;
	int bestDToF, bestSoftCost, proposedEvsInDummy, bestEvsInDummy;
	int numFeasMovesChain=0, numAccMovesChain=0, numProposedMovesChain=0;
	bool doingSwap, changed, doneKempe;
	
	//Structures used in the maxMatch algorithm and neighborhood operators
	IntVector c1Orig(numRooms), c2Orig(numRooms), d1Orig(numStudents), d2Orig(numStudents), timeslotForEvent(numEvents, -1), col, pred;
	IntQueue Q;
	//Structures used with Kempe chain interchanges
	vector<vector<int> > theCols(2,vector<int>(numRooms)), theGroups(2,vector<int>(numRooms));
	//Structures used with remove/replace procedures
	IntVector recCost;
	DoubleVector recTime;	

	//Make timeslotForEvent vector and copy of the current solution
	for(r=0; r<numRooms; r++){ for(c=0; c<NUMBEROFSLOTS; c++){if(theSolution.TT[r][c] != -1) timeslotForEvent[theSolution.TT[r][c]] = c;}}
	TTAndUnplaced bestSoFar = theSolution;

	//Calculate initial dummyCost (if we are not using one, this is set to zero). "dummyCost" is measured in the same units as dist. to feas. However, we can weight it later if needs be 
	calcDummyCost(theSolution,dummyCost,evsInDummy);
	
	//Calculate the time we have and use this to determine the SA run parameters
	clock_t totalTimeForSASoft = Part3ClockFinish - clock();
	
	determineInitTemp(initTemp,totalTimeForSASoft,timeslotForEvent,softCost,col,pred,Q,theSolution,studentavailability,stuDayCosts,dayCosts,theGroups,theCols,dummyCost,evsInDummy);
	bestDummyCost = dummyCost;
	bestEvsInDummy = evsInDummy;
	bestDToF = dToF;
	bestSoftCost = softCost;

	if(outputOn) cout << "Initial Temperature is = "<<initTemp<<endl;
	temperature = initTemp;

	//Set up some output streams. This gives us cost over time and per iteration for stage 2. (It is assumed that the optimisation starts at time 0)
	ofstream timeStream, itStream;
	timeStream.open("tEffort.txt"); itStream.open("cEffort.txt"); //Time is measured in seconds //Iterations is number of feasible-preserving moves
	timeStream<<bestSoftCost+bestDummyCost<<"\t0\n";
	itStream<<bestSoftCost+bestDummyCost<<"\t0\n";
	clock_t MCStart, MCEnd, MainSAClockStart = clock();

	//now do the SA
	while( Part3ClockFinish > clock() && (bestSoftCost+bestDummyCost) > 0){
		cnt = 0;
		numAccMovesChain = 0;
		numFeasMovesChain = 0;
		numProposedMovesChain = 0;

		//We time every 10th Markov chain and make adjustments to the cooling rate (making sure we hit the end temp at the timelimit)
		if(chainCnt%10==0) MCStart = clock();

		//Run the search procedure for 1 Marcov Chain at the current temperature
		while(cnt < mCLen && bestSoftCost+bestDummyCost > 0){
			
			//pick two cells to swap. First cell is an event, the second is a blank or an event
			pickAtomsSC(r1,c1,r2,c2,doingSwap,ev1,ev2,theSolution);
			propMoves++;
			numProposedMovesChain++;
						
			//See if this move maintains feasibility
			if(moveFeasible(r1,c1,r2,c2,timeslotForEvent,col,pred,Q,theSolution,c1Orig,c2Orig,changed,theCols,theGroups,doneKempe)){
			
				feasMoves++;
				numFeasMovesChain++;
				
				//It does maintain feasibility. We thus need to calculate the new score
				if(doneKempe) 
					recalculateStudentTimetablesKempe(c1,c2,d1Orig,d2Orig,studentavailability,stuDayCosts,dayCosts,theCols,theGroups);
				else 
					recalculateStudentTimetables(ev1,ev2,c1,c2,d1Orig,d2Orig,studentavailability,stuDayCosts,dayCosts);

				proposedSoftCost = totalCost(dayCosts,NUMDAYS);
				calcNewDummyCost(theSolution,c1,c2,c1Orig,c2Orig,proposedDummyCost,proposedEvsInDummy,dummyCost,evsInDummy);

				costChange = (weightCoefficient*proposedDummyCost + proposedSoftCost)-(weightCoefficient*dummyCost + softCost);
				
				if ( (costChange <= 0) || (rand()/double(RAND_MAX)) <= exp(-(costChange/temperature)) ){
					
					numAccMovesChain++;
					
					//The move has already been accepted so we just update the cost
					softCost = proposedSoftCost;
					dummyCost = proposedDummyCost;
					evsInDummy = proposedEvsInDummy;

					//if this is the best one we have found (ever) then we make a copy of it
					if (( (dToF+dummyCost) < (bestDToF + bestDummyCost) ) || ((dToF+dummyCost) == (bestDToF+bestDummyCost) && (softCost) < (bestSoftCost) ) ){
						bestSoFar = theSolution;
						bestSoftCost = softCost;
						bestDummyCost = dummyCost;
						bestEvsInDummy = evsInDummy;
						bestDToF = dToF;
						timeStream<<bestSoftCost+bestDummyCost<<"\t"<<(clock()-MainSAClockStart)/double(CLOCKS_PER_SEC)<<"\n";
						itStream<<bestSoftCost+bestDummyCost<<"\t"<<feasMoves<<"\n";
					}

				}
				else {
					//Neighbourhood move rejected
					if(changed)		resetMoveSC(c1,r1,c2,r2,c1Orig,c2Orig,timeslotForEvent,doingSwap,theSolution);
					if(doneKempe)	resetStudentTimetablesKempe(c1,c2,d1Orig,d2Orig,studentavailability,stuDayCosts,dayCosts,theCols,theGroups);
					else			resetStudentTimetables(ev2,ev1,c1,c2,d1Orig,d2Orig,studentavailability,stuDayCosts,dayCosts);
				}
			}
			else {
				//Move does not maintain feasibility, so reject
				if(changed) resetMoveSC(c1,r1,c2,r2,c1Orig,c2Orig,timeslotForEvent,doingSwap,theSolution);
			}

			cnt++;

		}
		//we have finished the currrent Marcov chain. If we've found a perfect solution then we exit, otherwise we update the temperature 
		if(chainCnt%10 == 0){
			//Every 10th Markov chain we refine the cooling rate alpha to make sure we hit the end temperature at the end of the run
			MCEnd = clock();
			MCTime = (MCEnd-MCStart)/double(CLOCKS_PER_SEC);
			timeRemaining = (Part3ClockFinish-MCEnd)/double(CLOCKS_PER_SEC);
			MCsRemaining = timeRemaining/MCTime;
			alpha = pow( (SAEndTemp/temperature),(1.0/MCsRemaining) );
			if(outputOn) cout<<"Chain "<<setw(6)<<chainCnt<<"/"<<chainCnt+int(MCsRemaining)<<". Cost = "<<setw(6)<<softCost<<" ["<<setw(6)<<dummyCost<<"] ("<<theSolution.Unplaced.size()<<","<<distToFeas(theSolution)<<") Best = "<<setw(6)<<bestSoftCost<<" ["<<setw(6)<<bestDummyCost<<"] ("<<bestSoFar.Unplaced.size()<<","<<bestDToF<<"). T = "<<setw(12)<<temperature<<". Feas = "<<setw(12)<<numFeasMovesChain/double(numProposedMovesChain)<<". Acc = "<<setw(12)<<numAccMovesChain/double(numFeasMovesChain)<<endl; 
		}
		
		chainCnt++;
		temperature = temperature*alpha;

	} //SA Algorithm Finished. Make sure we use the best solution found
	if(outputOn){
		cout<<"Chain "<<setw(6)<<chainCnt<<"/"<<chainCnt<<". Cost = "<<setw(6)<<softCost<<" ["<<setw(6)<<dummyCost<<"] ("<<theSolution.Unplaced.size()<<","<<distToFeas(theSolution)<<") Best = "<<setw(6)<<bestSoftCost<<" ["<<setw(6)<<bestDummyCost<<"] ("<<bestSoFar.Unplaced.size()<<","<<bestDToF<<"). T = "<<setw(12)<<temperature<<". Feas = "<<setw(12)<<numFeasMovesChain/double(numProposedMovesChain)<<". Acc = "<<setw(12)<<numAccMovesChain/double(numFeasMovesChain)<<endl; 
		cout<<"FINISHED SA NOW..."<<endl;
	}
	theSolution = bestSoFar;
	softCost = bestSoftCost;
	dToF = bestDToF;
	dummyCost = bestDummyCost;
	bestEvsInDummy = evsInDummy;

	timeStream<<"0\tX\t"<<(clock()-MainSAClockStart)/double(CLOCKS_PER_SEC)<<endl;
	itStream<<"0\tX\t"<<feasMoves<<endl;
	timeStream.close();
	itStream.close();
}


void doFeasRandWalk(int &softCost, clock_t &Part3ClockFinish, TTAndUnplaced &theSolution, 
				bool **studentavailability, int **stuDayCosts, int *dayCosts)
{
	//Simply performs a random walk (to gain an accurate measure for the feasibility ratio)
	if(outputOn) cout<<"Doing Random Walk";
	int r, c, r1, r2, c1, c2, ev1, ev2, cnt=0, bestSoftCost=softCost;
	bool doingSwap, changed, doneKempe;
	
	//Structures used in the maxMatch algorithm and neighborhood operators
	IntVector c1Orig(numRooms), c2Orig(numRooms), d1Orig(numStudents), d2Orig(numStudents), timeslotForEvent(numEvents, -1), col, pred;
	IntQueue Q;
	//Structures used with Kempe chain interchanges
	vector<vector<int> > theCols(2,vector<int>(numRooms)), theGroups(2,vector<int>(numRooms));

	//Make timeslotForEvent vector and copy of the current solution
	for(r=0; r<numRooms; r++){ for(c=0; c<NUMBEROFSLOTS; c++){if(theSolution.TT[r][c] != -1) timeslotForEvent[theSolution.TT[r][c]] = c;}}
	TTAndUnplaced bestSoFar = theSolution; 

	while( Part3ClockFinish > clock()){
		//pick two cells to swap. First cell is an event, the second is a blank or an event
		pickAtomsSC(r1,c1,r2,c2,doingSwap,ev1,ev2,theSolution);
		propMoves++;
		//See if this move maintains feasibility
		if(moveFeasible(r1,c1,r2,c2,timeslotForEvent,col,pred,Q,theSolution,c1Orig,c2Orig,changed,theCols,theGroups,doneKempe)){
			feasMoves++;
			//It does maintain feasibility. So we do the move and evalutate
			if(doneKempe) 
				recalculateStudentTimetablesKempe(c1,c2,d1Orig,d2Orig,studentavailability,stuDayCosts,dayCosts,theCols,theGroups);
			else 
				recalculateStudentTimetables(ev1,ev2,c1,c2,d1Orig,d2Orig,studentavailability,stuDayCosts,dayCosts);
			softCost = totalCost(dayCosts,NUMDAYS);
			//if this is the best one we have found (ever) then we make a copy of it
			if (softCost < bestSoftCost){
				bestSoFar = theSolution;
				bestSoftCost = softCost;
			}
		}
		else {
			//Move does not maintain feasibility, so reject
			if(changed) resetMoveSC(c1,r1,c2,r2,c1Orig,c2Orig,timeslotForEvent,doingSwap,theSolution);
		}
		cnt++;
		if(outputOn && cnt % 100000 == 0) cout<<".";
	} //Random Walk Finished. Make sure we use the best solution found
	if(outputOn)cout<<"\nFinished Random Walk. (Its =  "<<setw(6)<<cnt<<"; Best = "<<setw(6)<<bestSoftCost<<")"<<"Feas. Ratio = "<<feasMoves/double(propMoves)<<"\n"; 
	theSolution = bestSoFar;
	softCost = bestSoftCost;
}

void doFeasRandWalkPCol(TTAndUnplaced &theSolution, IntVector &timeslotForEvent, clock_t Part1ClockFinish)
{
	//Performs a random walk -- used in conjuction with partialCol algorithm
	int r1, r2, c1, c2, ev1, ev2, numFeasMoves=0, flexHolder=flexibility;
	bool doingSwap, changed, doneKempe;
	//Structures used in the maxMatch algorithm and neighborhood operators
	IntVector c1Orig(numRooms), c2Orig(numRooms), col, pred;
	IntQueue Q;
	//Structures used with Kempe chain interchanges
	vector<vector<int> > theCols(2,vector<int>(numRooms)), theGroups(2,vector<int>(numRooms));
	
	//Change flexibility so that Kempe chains and Max Match are used with this random walk operator
	flexibility = 3;
	//Do random walk until 100 feas moves have been performed. This gives > 95% chance that all timeslots will have undergone a move
	while(numFeasMoves < 100 && Part1ClockFinish < clock()){
		pickAtomsSC(r1,c1,r2,c2,doingSwap,ev1,ev2,theSolution);
		if(moveFeasible(r1,c1,r2,c2,timeslotForEvent,col,pred,Q,theSolution,c1Orig,c2Orig,changed,theCols,theGroups,doneKempe)){
			//Move has maintained feasibility.
			numFeasMoves++;
		}
		else {
			//Move does not maintain feasibility, so reset
			if(changed) resetMoveSC(c1,r1,c2,r2,c1Orig,c2Orig,timeslotForEvent,doingSwap,theSolution);
		}
	} 
	//Change flexibility back to it's original setting.
	flexibility = flexHolder;
}
